# ruff: noqa: PLR6104 RUF005
import unittest

from cubing_algs.algorithm import Algorithm
from cubing_algs.ergonomics import ErgonomicsData
from cubing_algs.exceptions import InvalidMoveError
from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.optimize import optimize_do_undo_moves
from cubing_algs.transform.optimize import optimize_double_moves
from cubing_algs.vcube import VCube


class AlgorithmTestCase(unittest.TestCase):

    def test_init_empty(self) -> None:
        algo = Algorithm()
        self.assertEqual(str(algo), '')

        algo.extend('R2 U')

        self.assertEqual(str(algo), 'R2 U')

    def test_init_value(self) -> None:
        algo = Algorithm([Move('F'), Move('R'), Move('U2')])
        self.assertEqual(str(algo), 'F R U2')

    def test_parse_moves(self) -> None:
        algo = Algorithm.parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo = Algorithm.parse_moves(['R2', 'U'])
        self.assertEqual(str(algo), 'R2 U')

        algo = Algorithm.parse_moves([Move('R2'), Move('U')])
        self.assertEqual(str(algo), 'R2 U')

        algo = Algorithm.parse_moves(Algorithm.parse_moves(['R2', 'U']))
        self.assertEqual(str(algo), 'R2 U')

    def test_parse_move(self) -> None:
        move = Algorithm.parse_move('R2')
        self.assertEqual(move, 'R2')

        move = Algorithm.parse_move(Move('R2'))
        self.assertEqual(move, 'R2')

        with self.assertRaises(InvalidMoveError):
            Algorithm.parse_move('R2 U')

    def test_append(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo.append('F2')
        self.assertEqual(str(algo), 'R2 U F2')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo.append(Move('D'))
        self.assertEqual(str(algo), 'R2 U F2 D')

        for m in algo:
            self.assertIsInstance(m, Move)

        with self.assertRaises(InvalidMoveError):
            algo.append('G')

        with self.assertRaises(InvalidMoveError):
            algo.append('F R')

        with self.assertRaises(InvalidMoveError):
            algo.append(Move('G'))

        with self.assertRaises(InvalidMoveError):
            algo.append(['F', 'R'])  # type: ignore[arg-type]

        with self.assertRaises(InvalidMoveError):
            algo.append([])  # type: ignore[arg-type]

    def test_extend(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo.extend(['F2', 'B'])
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo.extend([])
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo.extend('F R')
        self.assertEqual(str(algo), 'R2 U F2 B F R')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo.extend('D2')
        self.assertEqual(str(algo), 'R2 U F2 B F R D2')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo.extend([Move('L'), Move('B')])
        self.assertEqual(str(algo), 'R2 U F2 B F R D2 L B')

        for m in algo:
            self.assertIsInstance(m, Move)

        with self.assertRaises(InvalidMoveError):
            algo.extend(['F2', 'G'])

    def test_extend_with_algorithm(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo.extend(parse_moves('F2 B'))
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertIsInstance(m, Move)

    def test_add_operator(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo = algo + ['F2', 'B']
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo = algo + 'F R'
        self.assertEqual(str(algo), 'R2 U F2 B F R')

        for m in algo:
            self.assertIsInstance(m, Move)

        with self.assertRaises(InvalidMoveError):
            algo += 'F2 G'

    def test_iadd_operator(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo += ['F2', 'B']
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo += 'F R'
        self.assertEqual(str(algo), 'R2 U F2 B F R')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo += [Move('L2'), Move('D')]
        self.assertEqual(str(algo), 'R2 U F2 B F R L2 D')

        for m in algo:
            self.assertIsInstance(m, Move)

        with self.assertRaises(InvalidMoveError):
            algo + 'F2 G'

    def test_radd_operator(self) -> None:
        algo = 'F2R2' + parse_moves('D2 U')
        self.assertEqual(str(algo), 'F2 R2 D2 U')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo = ['R2', 'L'] + algo
        self.assertEqual(str(algo), 'R2 L F2 R2 D2 U')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo = [Move('D'), Move('U')] + algo
        self.assertEqual(str(algo), 'D U R2 L F2 R2 D2 U')

        for m in algo:
            self.assertIsInstance(m, Move)

        with self.assertRaises(InvalidMoveError):
            'F2 G' + algo

    def test_radd_operator_z(self) -> None:
        algo = 'z' + parse_moves('R2 U')
        self.assertEqual(str(algo), 'z R2 U')

    def test_radd_operator_zprime(self) -> None:
        algo = "z'" + parse_moves('R2 U')
        self.assertEqual(str(algo), "z' R2 U")

    def test_radd_operator_z2(self) -> None:
        algo = 'z2' + parse_moves('R2 U')
        self.assertEqual(str(algo), 'z2 R2 U')

    def test_add(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo += ['F2', 'B']
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo += [Move('D2'), Move('B2')]
        self.assertEqual(str(algo), 'R2 U F2 B D2 B2')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo += 'z y'
        self.assertEqual(str(algo), 'R2 U F2 B D2 B2 z y')

        for m in algo:
            self.assertIsInstance(m, Move)

        with self.assertRaises(InvalidMoveError):
            algo += 'F G'

    def test_add_exploded(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo += [*algo, 'F2', 'B']
        self.assertEqual(str(algo), 'R2 U R2 U F2 B')

        for m in algo:
            self.assertIsInstance(m, Move)

    def test_add_operator_with_algorithm(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo = algo + parse_moves('F2 B')
        self.assertEqual(str(algo), 'R2 U F2 B')

        for m in algo:
            self.assertIsInstance(m, Move)

    def test_insert(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        algo.insert(0, 'F2')
        self.assertEqual(str(algo), 'F2 R2 U')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo.insert(0, Move('L'))
        self.assertEqual(str(algo), 'L F2 R2 U')

        for m in algo:
            self.assertIsInstance(m, Move)

        with self.assertRaises(InvalidMoveError):
            algo.insert(0, 'G')

    def test_remove(self) -> None:
        algo = parse_moves('R2 U R2')
        self.assertEqual(str(algo), 'R2 U R2')

        algo.remove(Move('R2'))
        self.assertEqual(str(algo), 'U R2')

        for m in algo:
            self.assertIsInstance(m, Move)

    def test_remove_type_str(self) -> None:
        algo = parse_moves('R2 U R2')
        self.assertEqual(str(algo), 'R2 U R2')

        algo.remove('R2')  # type: ignore[arg-type]
        self.assertEqual(str(algo), 'U R2')

        for m in algo:
            self.assertIsInstance(m, Move)

    def test_pop(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')
        popped = algo.pop()
        self.assertEqual(str(algo), 'R2')
        self.assertEqual(popped, 'U')

        for m in algo:
            self.assertIsInstance(m, Move)

    def test_copy(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(str(algo), 'R2 U')

        copy = algo.copy()
        self.assertIsInstance(copy, Algorithm)
        self.assertEqual(str(copy), 'R2 U')

        for m in algo:
            self.assertIsInstance(m, Move)
        for m in copy:
            self.assertIsInstance(m, Move)

        algo.pop()
        self.assertEqual(str(algo), 'R2')
        self.assertEqual(str(copy), 'R2 U')

        for m in algo:
            self.assertIsInstance(m, Move)
        for m in copy:
            self.assertIsInstance(m, Move)

    def test_iter(self) -> None:
        algo = parse_moves('R2 U')
        for m, n in zip(algo, ['R2', 'U'], strict=True):
            self.assertEqual(m, n)

    def test_getitem(self) -> None:
        algo = parse_moves('R2 U')
        self.assertEqual(algo[1], 'U')
        self.assertIsInstance(algo[1], Move)

        algo = parse_moves('R2 U F D B L')[1:4]
        self.assertEqual(str(algo), 'U F D')
        self.assertIsInstance(algo, Algorithm)
        self.assertIsInstance(algo[1], Move)

    def test_setitem_slice(self) -> None:
        algo = parse_moves('R2 U F')
        algo[2:] = []
        self.assertEqual(str(algo), 'R2 U')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo = parse_moves('R2 U F')
        algo[1:] = []
        self.assertEqual(str(algo), 'R2')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo = parse_moves('R2 U F')
        new_algo = parse_moves('B2 D')
        algo[2:] = new_algo
        self.assertEqual(str(algo), 'R2 U B2 D')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo = parse_moves('R2 U F')
        new_algo = parse_moves('B2 D')
        algo[1:] = new_algo
        self.assertEqual(str(algo), 'R2 B2 D')

        for m in algo:
            self.assertIsInstance(m, Move)

        algo = parse_moves('R2 U F L D')
        new_algo = parse_moves('B2 D')
        algo[1:3] = new_algo
        self.assertEqual(str(algo), 'R2 B2 D L D')

        for m in algo:
            self.assertIsInstance(m, Move)

    def test_setitem(self) -> None:
        algo = parse_moves('R2 U')
        algo[1] = Move('B')
        self.assertEqual(str(algo), 'R2 B')

        for m in algo:
            self.assertIsInstance(m, Move)

    def test_delitem(self) -> None:
        algo = parse_moves('R2 U')
        del algo[1]
        self.assertEqual(str(algo), 'R2')

        for m in algo:
            self.assertIsInstance(m, Move)

    def test_contains(self) -> None:
        algo = parse_moves('R2 U')
        self.assertIn(Move('U'), algo)
        self.assertIn(Move('R2'), algo)
        self.assertNotIn(Move('R'), algo)

    def test_contains_type_str(self) -> None:
        algo = parse_moves('R2 U')
        self.assertIn('U', algo)
        self.assertIn('R2', algo)
        self.assertNotIn('R', algo)
        self.assertNotIn('2', algo)

    def test_count(self) -> None:
        algo = parse_moves('R2 U F R R2')
        self.assertEqual(algo.count(Move('R')), 1)
        self.assertEqual(algo.count(Move('R2')), 2)
        self.assertEqual(algo.count(Move('L')), 0)

    def test_count_type_str(self) -> None:
        algo = parse_moves('R2 U F R R2')
        self.assertEqual(algo.count('R'), 1)   # type: ignore[arg-type]
        self.assertEqual(algo.count('R2'), 2)  # type: ignore[arg-type]
        self.assertEqual(algo.count('L'), 0)   # type: ignore[arg-type]
        self.assertEqual(algo.count('2'), 0)   # type: ignore[arg-type]

    def test_index(self) -> None:
        algo = parse_moves('R2 U F R R2')
        self.assertEqual(algo.index(Move('R')), 3)
        self.assertEqual(algo.index(Move('R2')), 0)

        with self.assertRaises(ValueError):
            algo.index(Move('L'))

    def test_index_type_str(self) -> None:
        algo = parse_moves('R2 U F R R2')
        self.assertEqual(algo.index('R'), 3)   # type: ignore[arg-type]
        self.assertEqual(algo.index('R2'), 0)  # type: ignore[arg-type]

        with self.assertRaises(ValueError):
            algo.index('L')  # type: ignore[arg-type]

        with self.assertRaises(ValueError):
            algo.index('2')  # type: ignore[arg-type]

    def test_length(self) -> None:
        algo = parse_moves('R2 U')

        self.assertEqual(len(algo), 2)

    def test_str(self) -> None:
        algo = parse_moves('R2 U')

        self.assertEqual(str(algo), 'R2 U')

    def test_repr(self) -> None:
        algo = parse_moves('R2 U')

        self.assertEqual(repr(algo), 'Algorithm("R2U")')

    def test_eq(self) -> None:
        algo = parse_moves('R2 U')
        algo_bis = parse_moves('R2 U')

        self.assertEqual(algo, algo_bis)

    def test_eq_copy(self) -> None:
        algo = parse_moves('R2 U')
        algo_copy = algo.copy()

        self.assertEqual(algo, algo_copy)

    def test_eq_list(self) -> None:
        algo = parse_moves('R2 U')
        algo_list = [Move('R2'), Move('U')]

        self.assertEqual(algo, algo_list)

    def test_transform(self) -> None:
        algo = parse_moves('R R U F2 F2')
        expected = parse_moves('R2 U')

        self.assertEqual(
            algo.transform(
                optimize_do_undo_moves,
                optimize_double_moves,
            ),
            expected,
        )

        algo = parse_moves('R U F2')
        expected = parse_moves('R U F2')

        self.assertEqual(
            algo.transform(
                optimize_do_undo_moves,
                optimize_double_moves,
            ),
            expected,
        )

    def test_transform_to_fixpoint(self) -> None:
        algo = parse_moves("R R F F' R2 U F2")
        expected = parse_moves('R2 R2 U F2')

        self.assertEqual(
            algo.transform(
                optimize_do_undo_moves,
                optimize_double_moves,
            ),
            expected,
        )

        algo = parse_moves("R R F F' R2 U F2")
        expected = parse_moves('U F2')

        self.assertEqual(
            algo.transform(
                optimize_do_undo_moves,
                optimize_double_moves,
                to_fixpoint=True,
            ),
            expected,
        )

    def test_min_cube_size(self) -> None:
        algo = parse_moves("B' R2 U F2")

        self.assertEqual(
            algo.min_cube_size,
            2,
        )

        algo = parse_moves("B' R2 M U F2")

        self.assertEqual(
            algo.min_cube_size,
            3,
        )

        algo = parse_moves("B' r2 U F2")

        self.assertEqual(
            algo.min_cube_size,
            3,
        )

        algo = parse_moves("B' R2 U Fw2")

        self.assertEqual(
            algo.min_cube_size,
            3,
        )

        algo = parse_moves("B' R2 U 2Fw2")

        self.assertEqual(
            algo.min_cube_size,
            3,
        )

        algo = parse_moves("B' R2 U 3Fw2")

        self.assertEqual(
            algo.min_cube_size,
            6,
        )

        algo = parse_moves("B' R2 U 2-3Fw2")

        self.assertEqual(
            algo.min_cube_size,
            6,
        )

        algo = parse_moves("B' R2 U 4Fw2")

        self.assertEqual(
            algo.min_cube_size,
            8,
        )

        algo = parse_moves("B' R2 U 2-4Fw2")

        self.assertEqual(
            algo.min_cube_size,
            8,
        )

    def test_is_standard(self) -> None:
        algo = parse_moves("B' R2 U 2-4Fw2")

        self.assertTrue(algo.is_standard)
        self.assertFalse(algo.is_sign)

    def test_is_sign(self) -> None:
        algo = parse_moves("B' R2 U 2-4f2")

        self.assertTrue(algo.is_sign)
        self.assertFalse(algo.is_standard)

    def test_has_rotations(self) -> None:
        algo = parse_moves("B' R2 U 2-4Fw2")

        self.assertTrue(algo.has_rotations)

        algo = parse_moves("B' R2 U 2-4f2")

        self.assertTrue(algo.has_rotations)

        algo = parse_moves("B' R2 U x")

        self.assertTrue(algo.has_rotations)

        algo = parse_moves('R2 E U')

        self.assertTrue(algo.has_rotations)

        algo = parse_moves('R2 F U D2 L B')

        self.assertFalse(algo.has_rotations)

    def test_has_internal_rotations(self) -> None:
        algo = parse_moves("B' R2 U 2-4Fw2")

        self.assertTrue(algo.has_internal_rotations)

        algo = parse_moves("B' R2 U 2-4f2")

        self.assertTrue(algo.has_internal_rotations)

        algo = parse_moves('R2 E U')

        self.assertTrue(algo.has_internal_rotations)

        algo = parse_moves("B' R2 U x")

        self.assertFalse(algo.has_internal_rotations)


class AlgorithmCyclesPropertyTestCase(unittest.TestCase):
    """Test cases for the Algorithm.cycles property."""

    def test_empty_algorithm_cycles(self) -> None:
        """Test cycles property for empty algorithm."""
        algo = Algorithm()
        result = algo.cycles
        self.assertEqual(result, 0)

    def test_single_move_cycles(self) -> None:
        """Test cycles property for single move."""
        algo = Algorithm.parse_moves('R')
        result = algo.cycles
        self.assertEqual(result, 4)  # R has order 4

    def test_sexy_move_cycles(self) -> None:
        """Test cycles property for sexy move."""
        algo = Algorithm.parse_moves("R U R' U'")
        result = algo.cycles
        self.assertEqual(result, 6)  # Known order of sexy move

    def test_half_turn_cycles(self) -> None:
        """Test cycles property for half turn."""
        algo = Algorithm.parse_moves('R2')
        result = algo.cycles
        self.assertEqual(result, 2)  # R2 has order 2

    def test_identity_cycles(self) -> None:
        """Test cycles property for identity algorithm."""
        algo = Algorithm.parse_moves("R R'")
        result = algo.cycles
        self.assertEqual(result, 1)

    def test_complex_algorithm_cycles(self) -> None:
        """Test cycles property for complex algorithm."""
        algo = Algorithm.parse_moves("R U2 R' D' R U' R' D")
        result = algo.cycles
        self.assertIsInstance(result, int)
        self.assertGreaterEqual(result, 0)
        self.assertLess(result, 100)

    def test_cycles_return_int_non_negative(self) -> None:
        """Test that cycles property returns integer."""
        test_cases = ['R', "R U R' U'", 'F2', 'x', 'M', '']

        for moves_str in test_cases:
            with self.subTest(moves=moves_str):
                if not moves_str:
                    algo = Algorithm()
                else:
                    algo = Algorithm.parse_moves(moves_str)
                result = algo.cycles
                self.assertIsInstance(result, int)
                self.assertGreaterEqual(result, 0)

    def test_cycles_with_rotations(self) -> None:
        """Test cycles property with cube rotations."""
        algo = Algorithm.parse_moves('x y z')
        result = algo.cycles
        self.assertIsInstance(result, int)
        self.assertEqual(result, 1)

    def test_cycles_with_slice_moves(self) -> None:
        """Test cycles property with slice moves."""
        algo = Algorithm.parse_moves('M E S')
        result = algo.cycles
        self.assertIsInstance(result, int)
        self.assertEqual(result, 4)

    def test_cycles_with_wide_moves(self) -> None:
        """Test cycles property with wide moves."""
        algo = Algorithm.parse_moves('r u f')
        result = algo.cycles
        self.assertIsInstance(result, int)
        self.assertEqual(result, 70)

    def test_cycles_with_pauses(self) -> None:
        """Test cycles property with pauses."""
        algo = Algorithm.parse_moves('R U . F')
        result = algo.cycles
        self.assertIsInstance(result, int)
        self.assertEqual(result, 80)

    def test_cycles_with_timed_moves(self) -> None:
        """Test cycles property with timed moves."""
        algo = Algorithm.parse_moves('R@50 U@75 F@100')
        result = algo.cycles
        self.assertIsInstance(result, int)
        self.assertEqual(result, 80)


class AlgorithmShowTestCase(unittest.TestCase):
    """Test cases for the Algorithm.show method."""

    def test_show_method_basic(self) -> None:
        """Test show method returns VCube instance."""
        algo = Algorithm.parse_moves("R U R'")
        result = algo.show()

        self.assertIsInstance(result, VCube)

    def test_show_method_with_parameters(self) -> None:
        """Test show method with mode and orientation parameters."""
        algo = Algorithm.parse_moves("R U R'")
        result = algo.show(mode='oll', orientation='FU')

        self.assertIsInstance(result, VCube)

    def test_show_method_empty_algorithm(self) -> None:
        """Test show method with empty algorithm."""
        algo = Algorithm()
        result = algo.show()

        self.assertIsInstance(result, VCube)


class AlgorithmErgonomicsTestCase(unittest.TestCase):
    """Test cases for the Algorithm.ergonomics property."""

    def test_ergonomics_property_returns_ergonomics_data(self) -> None:
        """Test ergonomics property returns ErgonomicsData."""
        algo = Algorithm.parse_moves("R U R' U'")
        ergo = algo.ergonomics

        self.assertIsInstance(ergo, ErgonomicsData)

    def test_ergonomics_property_has_expected_fields(self) -> None:
        """Test ergonomics property returns data with expected fields."""
        algo = Algorithm.parse_moves("R U R' U'")
        ergo = algo.ergonomics

        # Check that all expected fields are present
        self.assertIsNotNone(ergo.comfort_score)
        self.assertIsNotNone(ergo.ergonomic_rating)
        self.assertIsNotNone(ergo.hand_balance_ratio)

    def test_ergonomics_property_empty_algorithm(self) -> None:
        """Test ergonomics property with empty algorithm."""
        algo = Algorithm()
        ergo = algo.ergonomics

        self.assertIsInstance(ergo, ErgonomicsData)



import unittest
from io import StringIO
from typing import Any
from unittest.mock import patch

from cubing_algs.constants import FACES
from cubing_algs.constants import INITIAL_STATE
from cubing_algs.exceptions import InvalidCubeStateError
from cubing_algs.exceptions import InvalidFaceError
from cubing_algs.exceptions import InvalidMoveError
from cubing_algs.integrity import VCubeIntegrityChecker
from cubing_algs.masks import F2L_MASK
from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.wide import unwide_rotation_moves
from cubing_algs.vcube import VCube


class VCubeTestCase(unittest.TestCase):
    maxDiff = None

    def test_state(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.state,
            INITIAL_STATE,
        )

        result = cube.rotate('R2 U2')
        self.assertEqual(
            result,
            'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFBB',
        )

        self.assertEqual(
            result,
            cube.state,
        )

    def test_is_solved(self) -> None:
        cube = VCube()

        self.assertTrue(
            cube.is_solved,
        )

        cube.rotate('R2 U2')
        self.assertFalse(
            cube.is_solved,
        )

    def test_is_solved_oriented(self) -> None:
        cube = VCube()
        cube.rotate('z2')

        self.assertTrue(cube.is_solved)

    def test_rotate_history(self) -> None:
        cube = VCube()
        cube.rotate('R')

        self.assertEqual(cube.history, ['R'])

        cube.rotate('L', history=False)

        self.assertEqual(cube.history, ['R'])

    def test_rotate_move_history(self) -> None:
        cube = VCube()
        cube.rotate_move('R')

        self.assertEqual(cube.history, ['R'])

        cube.rotate_move('L', history=False)

        self.assertEqual(cube.history, ['R'])

    def test_copy(self) -> None:
        cube = VCube()
        cube.rotate('R2 F2 D2 B')
        copy = cube.copy()

        self.assertEqual(
            cube.state,
            copy.state,
        )
        self.assertFalse(copy.history)

    def test_full_copy(self) -> None:
        cube = VCube()
        cube.rotate('R2 F2 D2 B')
        copy = cube.copy(full=True)

        self.assertEqual(
            cube.state,
            copy.state,
        )
        self.assertTrue(copy.history)

    def test_from_cubies(self) -> None:
        cp = [0, 5, 2, 1, 7, 4, 6, 3]
        co = [1, 2, 0, 2, 1, 1, 0, 2]
        ep = [1, 9, 2, 3, 11, 8, 6, 7, 4, 5, 10, 0]
        eo = [1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]
        so = [0, 1, 2, 3, 4, 5]
        facelets = 'UUFUUFLLFUUURRRRRRFFRFFDFFDRRBDDBDDBLLDLLDLLDLBBUBBUBB'

        cube = VCube.from_cubies(cp, co, ep, eo, so)
        self.assertEqual(cube.state, facelets)

        cube = VCube()
        cube.rotate('F R')

        self.assertEqual(cube.state, facelets)

    def test_from_cubies_scheme(self) -> None:
        cp = [0, 5, 2, 1, 7, 4, 6, 3]
        co = [1, 2, 0, 2, 1, 1, 0, 2]
        ep = [1, 9, 2, 3, 11, 8, 6, 7, 4, 5, 10, 0]
        eo = [1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]
        so = [0, 1, 2, 3, 4, 5]
        facelets = '111111011111011011011010010010001001110110000111111100'

        cube = VCube.from_cubies(
            cp, co, ep, eo, so,
            F2L_MASK,
        )
        self.assertEqual(cube.state, facelets)

        cube = VCube(F2L_MASK, check=False)
        cube.rotate('F R')

        self.assertEqual(cube.state, facelets)

    def test_to_cubies(self) -> None:
        cp = [0, 5, 2, 1, 7, 4, 6, 3]
        co = [1, 2, 0, 2, 1, 1, 0, 2]
        ep = [1, 9, 2, 3, 11, 8, 6, 7, 4, 5, 10, 0]
        eo = [1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]
        so = [0, 1, 2, 3, 4, 5]
        facelets = 'UUFUUFLLFUUURRRRRRFFRFFDFFDRRBDDBDDBLLDLLDLLDLBBUBBUBB'

        self.assertEqual(
            VCube(facelets).to_cubies,
            (
                cp, co,
                ep, eo,
                so,
            ),
        )

    def test_from_cubies_equality(self) -> None:
        cube = VCube()
        cube.rotate('F R')
        n_cube = VCube.from_cubies(*cube.to_cubies)

        self.assertEqual(
            cube.state,
            n_cube.state,
        )

    def test_from_cubies_oriented_equality(self) -> None:
        cube = VCube()
        cube.rotate('F R x')
        n_cube = VCube.from_cubies(*cube.to_cubies)

        self.assertEqual(
            cube.state,
            n_cube.state,
        )

    def test_display(self) -> None:
        cube = VCube()
        cube.rotate('F R U')

        result = cube.display()

        lines = [line for line in result.split('\n') if line.strip()]

        self.assertEqual(len(lines), 9)
        self.assertEqual(len(cube.history), 3)

    def test_display_orientation_restore(self) -> None:
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(len(cube.history), 3)

        state = cube.state

        cube.display(orientation='DF')

        self.assertEqual(len(cube.history), 3)
        self.assertEqual(state, cube.state)

    def test_display_orientation_different(self) -> None:
        cube_1 = VCube()
        cube_2 = VCube()

        view_1 = cube_1.display()
        view_2 = cube_2.display(orientation='DF')

        self.assertNotEqual(view_1, view_2)

    def test_get_face(self) -> None:
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face('U'),
            'LUULUUFFF',
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face('U'),
            'BDDBDDBRR',
        )

    def test_get_face_by_center(self) -> None:
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face_by_center('U'),
            'LUULUUFFF',
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face_by_center('U'),
            'FFFUULUUL',
        )

    def test_get_face_center(self) -> None:
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face_by_center('U'),
            'LUULUUFFF',
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face_by_center('U'),
            'FFFUULUUL',
        )

    def test_get_face_index(self) -> None:
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face_index('U'),
            0,
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face_index('U'),
            3,
        )

    def test_get_face_center_indexes(self) -> None:
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face_center_indexes(),
            ['U', 'R', 'F', 'D', 'L', 'B'],
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face_center_indexes(),
            ['D', 'L', 'F', 'U', 'R', 'B'],
        )

    def test_str(self) -> None:
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            str(cube),
            'U: LUULUUFFF\n'
            'R: LBBRRRRRR\n'
            'F: UUUFFDFFD\n'
            'D: RRBDDBDDB\n'
            'L: FFRLLDLLD\n'
            'B: LLDUBBUBB',
        )

    def test_repr(self) -> None:
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            repr(cube),
            "VCube('LUULUUFFFLBBRRRRRRUUUFFDFFDRRBDDBDDBFFRLLDLLDLLDUBBUBB')",
        )


class VCubeOrientedCopyTestCase(unittest.TestCase):
    maxDiff = None

    def test_oriented_copy_faces(self) -> None:
        cube = VCube()

        self.assertNotEqual(
            cube.state,
            cube.oriented_copy('DF').state,
        )

    def test_oriented_copy_top_only(self) -> None:
        cube = VCube()

        self.assertNotEqual(
            cube.state,
            cube.oriented_copy('D').state,
        )

    def test_oriented_copy_faces_stable(self) -> None:
        cube = VCube()
        base_state = cube.state
        cube.oriented_copy('UF')

        self.assertEqual(
            cube.state,
            base_state,
        )

    def test_oriented_copy_invalid_empty(self) -> None:
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('')

    def test_oriented_copy_invalid_too_much(self) -> None:
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('FRU')

    def test_oriented_copy_invalid_top_face(self) -> None:
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('TF')

    def test_oriented_copy_invalid_front_face(self) -> None:
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('FT')

    def test_oriented_copy_invalid_opposite_face(self) -> None:
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('FB')

    def test_oriented_copy_history_preservation(self) -> None:
        cube = VCube()
        cube.rotate('R F')

        self.assertEqual(
            len(cube.history),
            2,
        )

        oriented = cube.oriented_copy('DF')

        self.assertEqual(
            len(cube.history),
            2,
        )

        self.assertEqual(
            len(oriented.history),
            0,
        )

    def test_oriented_copy_history_tracking(self) -> None:
        cube = VCube()
        cube.rotate('R F')

        oriented = cube.oriented_copy('DF', full=True)

        self.assertEqual(
            oriented.history,
            ['R', 'F', 'z2'],
        )

        oriented = cube.oriented_copy('DR', full=True)

        self.assertEqual(
            oriented.history,
            ['R', 'F', 'y', 'z2'],
        )

    def test_all_edge_reorientation(self) -> None:
        orientations = [
            'UF', 'UB', 'UR', 'UL',
            'DF', 'DB', 'DR', 'DL',
            'FU', 'FD', 'FR', 'FL',
            'BU', 'BD', 'BR', 'BL',
            'RU', 'RD', 'RF', 'RB',
            'LU', 'LD', 'LF', 'LB',
        ]

        for orientation in orientations:
            with self.subTest(orientation=orientation):
                cube = VCube().oriented_copy(orientation)

                self.assertEqual(
                    cube.state[4],
                    orientation[0],
                )

                self.assertEqual(
                    cube.state[21],
                    orientation[1],
                )

    def test_all_reorientation(self) -> None:
        orientations = [
            'U', 'R', 'F', 'D', 'L', 'B',
        ]

        for orientation in orientations:
            with self.subTest(orientation=orientation):
                cube = VCube().oriented_copy(orientation)

                self.assertEqual(
                    cube.state[4],
                    orientation[0],
                )


class VCubeCheckIntegrityTestCase(unittest.TestCase):
    """Tests pour les nouvelles vérifications de check_integrity"""

    def test_initial(self) -> None:
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFBB'

        cube = VCube(initial)

        self.assertEqual(
            cube.state,
            initial,
        )

    def test_get_face_center_indexes_not_implemented(self) -> None:
        """Test that get_face_center_indexes raises NotImplementedError."""
        class IncompleteVCube(VCubeIntegrityChecker):
            """Incomplete implementation for testing."""
            size = 3
            face_size = 9
            face_number = 6
            _state = 'UUUUUUUUURRRRRRRRRFFFFFFFFFDDDDDDDDDLLLLLLLLLBBBBBBBBB'

        incomplete_cube = IncompleteVCube()

        with self.assertRaises(NotImplementedError):
            incomplete_cube.get_face_center_indexes()

    def test_invalid_length_no_check(self) -> None:
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFB'

        cube = VCube(initial, check=False)
        self.assertEqual(cube.state, initial)

    def test_invalid_length(self) -> None:
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFB'

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'State string must be 54 characters long',
        ):
            VCube(initial)

    def test_invalid_character(self) -> None:
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFBT'

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'State string can only contains U R F D L B characters',
        ):
            VCube(initial)

    def test_invalid_face(self) -> None:
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFBF'

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'State string must have 9 of each color',
        ):
            VCube(initial)

    def test_invalid_centers_not_unique(self) -> None:
        invalid_state = (
            'UUUUUUUUR'
            'RRRRURRRR'
            'FFFFFFFFF'
            'DDDDDDDDD'
            'LLLLLLLLL'
            'BBBBBBBBB'
        )

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Face centers must be unique',
        ):
            VCube(invalid_state)

    def test_invalid_corner_orientation_sum(self) -> None:
        co = [1, 0, 0, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Sum of corner orientations must be divisible by 3',
        ):
            VCube().check_corner_sum(co)

    def test_invalid_edge_orientation_sum(self) -> None:
        eo = [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Sum of edge orientations must be even',
        ):
            VCube().check_edge_sum(eo)

    def test_invalid_corner_permutation_duplicate(self) -> None:
        cp = [0, 0, 2, 3, 4, 5, 6, 7]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner permutation must contain exactly '
                'one instance of each corner',
        ):
            VCube().check_corner_permutations(cp)

    def test_invalid_edge_permutation_duplicate(self) -> None:
        ep = [0, 0, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Edge permutation must contain exactly '
                'one instance of each edge',
        ):
            VCube().check_edge_permutations(ep)

    def test_invalid_corner_orientation_value(self) -> None:
        co = [3, 0, 0, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner orientation must be 0, 1, or 2 '
                'for each corner',
        ):
            VCube().check_corner_orientations(co)

    def test_invalid_edge_orientation_value(self) -> None:
        eo = [2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Edge orientation must be 0 or 1 '
                'for each edge',
        ):
            VCube().check_edge_orientations(eo)

    def test_invalid_center_orientation_value(self) -> None:
        so = [7, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Center orientation must be between 0 and 5 '
                'for each center',
        ):
            VCube().check_center_orientations(so)

    def test_invalid_permutation_parity(self) -> None:
        # Swap 0,1 = 1 inversion (odd)
        cp = [1, 0, 2, 3, 4, 5, 6, 7]
        # Identity = 0 inversions (even)
        ep = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner and edge permutation parities must be equal',
        ):
            VCube().check_permutation_parity(cp, ep)

    @unittest.mock.patch.object(VCube, 'check_colors')
    def test_invalid_corner_same_colors(self, *_: Any) -> None:
        invalid_state_list = list(INITIAL_STATE)
        # Corner URF: same color on the 2 faces
        invalid_state_list[8] = invalid_state_list[9]
        invalid_state = ''.join(invalid_state_list)

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner 0 must have 3 different colors, got',
        ):
            VCube(invalid_state)

    @unittest.mock.patch.object(VCube, 'check_colors')
    def test_invalid_edge_same_colors(self, *_: Any) -> None:
        invalid_state_list = list(INITIAL_STATE)
        # Edge UR: same color on the 2 faces
        invalid_state_list[5] = invalid_state_list[10]
        invalid_state = ''.join(invalid_state_list)

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Edge 0 must have 2 different colors, got ',
        ):
            VCube(invalid_state)

    @unittest.mock.patch.object(VCube, 'check_colors')
    def test_invalid_corner_opposite_colors(self, *_: Any) -> None:
        invalid_state_list = list(INITIAL_STATE)
        invalid_state_list[8] = 'U'  # Face U
        invalid_state_list[9] = 'D'  # Opposite face D
        invalid_state_list[20] = 'F'  # Third face
        invalid_state = ''.join(invalid_state_list)

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner 0 cannot have opposite colors '
                'U and D',
        ):
            VCube(invalid_state)

    @unittest.mock.patch.object(VCube, 'check_colors')
    def test_invalid_edge_opposite_colors(self, *_: Any) -> None:
        invalid_state_list = list(INITIAL_STATE)
        invalid_state_list[5] = 'F'
        invalid_state_list[10] = 'B'  # Opposite color
        invalid_state = ''.join(invalid_state_list)

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Edge 0 cannot have opposite colors '
                'F and B',
        ):
            VCube(invalid_state)

    def test_valid_complex_scramble(self) -> None:
        cube = VCube()
        complex_scramble = (
            "R U2 R' D' R U' R' D R' U "
            "R U' R' U R U2 R' U' R U' R'"
        )
        cube.rotate(complex_scramble)

        self.assertTrue(cube.check_integrity())

    def test_rotations_preserve_validity(self) -> None:
        cube = VCube()
        rotations = ['x', 'y', 'z', "x'", "y'", "z'", 'x2', 'y2', 'z2']

        for rotation in rotations:
            with self.subTest(rotation=rotation):
                cube_copy = cube.copy()
                cube_copy.rotate(rotation)
                self.assertTrue(cube_copy.check_integrity())

    def test_preserve_validity(self) -> None:
        cube = VCube()

        self.assertTrue(cube.check_integrity())

    def test_oriented_preserve_validity(self) -> None:
        cube = VCube()
        cube.rotate('z2')

        self.assertTrue(
            VCube(cube.state).check_integrity(),
        )


class VCubeRotateTestCase(unittest.TestCase):

    def test_rotate_types(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate(parse_moves('R F') + 'z2'),
            'BDDBDDRRRBLLDLLDLLDDDFFFFFFLLLFUUFUURRFRRURRUBBUBBUBBU',
        )

        cube = VCube()

        self.assertEqual(
            cube.rotate('z' + parse_moves('R F')),
            'LLFLLFDDDLUULUUFUUFFFFFFRRRUUURRBRRBDDRDDRDDBLBBLBBLBB',
        )

        cube = VCube()

        self.assertEqual(
            cube.rotate('z2' + parse_moves('R F')),
            'DDFDDFRRRDLLDLLFLLFFFFFFUUULLLUUBUUBRRURRURRBDBBDBBDBB',
        )

    def test_rotate_typing(self) -> None:
        expected = 'UUFUUFUUFRRRRRRRRRFFDFFDFFDDDBDDBDDBLLLLLLLLLUBBUBBUBB'

        move_str = 'R'
        cube = VCube()
        cube.rotate(move_str)
        self.assertEqual(cube.state, expected)

        move_algo = parse_moves('R')
        cube = VCube()
        cube.rotate(move_algo)
        self.assertEqual(cube.state, expected)

        move_move = Move('R')
        cube = VCube()
        cube.rotate(move_move)
        self.assertEqual(cube.state, expected)

    def test_rotate_u(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('U'),
            'UUUUUUUUUBBBRRRRRRRRRFFFFFFDDDDDDDDDFFFLLLLLLLLLBBBBBB',
        )

        self.assertEqual(
            cube.rotate("U'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('U2'),
            'UUUUUUUUULLLRRRRRRBBBFFFFFFDDDDDDDDDRRRLLLLLLFFFBBBBBB',
        )

    def test_rotate_r(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('R'),
            'UUFUUFUUFRRRRRRRRRFFDFFDFFDDDBDDBDDBLLLLLLLLLUBBUBBUBB',
        )

        self.assertEqual(
            cube.rotate("R'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('R2'),
            'UUDUUDUUDRRRRRRRRRFFBFFBFFBDDUDDUDDULLLLLLLLLFBBFBBFBB',
        )

    def test_rotate_f(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('F'),
            'UUUUUULLLURRURRURRFFFFFFFFFRRRDDDDDDLLDLLDLLDBBBBBBBBB',
        )

        self.assertEqual(
            cube.rotate("F'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('F2'),
            'UUUUUUDDDLRRLRRLRRFFFFFFFFFUUUDDDDDDLLRLLRLLRBBBBBBBBB',
        )

    def test_rotate_d(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('D'),
            'UUUUUUUUURRRRRRFFFFFFFFFLLLDDDDDDDDDLLLLLLBBBBBBBBBRRR',
        )

        self.assertEqual(
            cube.rotate("D'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('D2'),
            'UUUUUUUUURRRRRRLLLFFFFFFBBBDDDDDDDDDLLLLLLRRRBBBBBBFFF',
        )

    def test_rotate_l(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('L'),
            'BUUBUUBUURRRRRRRRRUFFUFFUFFFDDFDDFDDLLLLLLLLLBBDBBDBBD',
        )

        self.assertEqual(
            cube.rotate("L'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('L2'),
            'DUUDUUDUURRRRRRRRRBFFBFFBFFUDDUDDUDDLLLLLLLLLBBFBBFBBF',
        )

    def test_rotate_b(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('B'),
            'RRRUUUUUURRDRRDRRDFFFFFFFFFDDDDDDLLLULLULLULLBBBBBBBBB',
        )

        self.assertEqual(
            cube.rotate("B'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('B2'),
            'DDDUUUUUURRLRRLRRLFFFFFFFFFDDDDDDUUURLLRLLRLLBBBBBBBBB',
        )

    def test_rotate_m(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('M'),
            'UBUUBUUBURRRRRRRRRFUFFUFFUFDFDDFDDFDLLLLLLLLLBDBBDBBDB',
        )

        self.assertEqual(
            cube.rotate("M'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('M2'),
            'UDUUDUUDURRRRRRRRRFBFFBFFBFDUDDUDDUDLLLLLLLLLBFBBFBBFB',
        )

    def test_rotate_s(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('S'),
            'UUULLLUUURURRURRURFFFFFFFFFDDDRRRDDDLDLLDLLDLBBBBBBBBB',
        )

        self.assertEqual(
            cube.rotate("S'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('S2'),
            'UUUDDDUUURLRRLRRLRFFFFFFFFFDDDUUUDDDLRLLRLLRLBBBBBBBBB',
        )

    def test_rotate_e(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('E'),
            'UUUUUUUUURRRFFFRRRFFFLLLFFFDDDDDDDDDLLLBBBLLLBBBRRRBBB',
        )

        self.assertEqual(
            cube.rotate("E'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('E2'),
            'UUUUUUUUURRRLLLRRRFFFBBBFFFDDDDDDDDDLLLRRRLLLBBBFFFBBB',
        )

    def test_rotate_x(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('x'),
            'FFFFFFFFFRRRRRRRRRDDDDDDDDDBBBBBBBBBLLLLLLLLLUUUUUUUUU',
        )

        self.assertEqual(
            cube.rotate("x'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('x2'),
            'DDDDDDDDDRRRRRRRRRBBBBBBBBBUUUUUUUUULLLLLLLLLFFFFFFFFF',
        )

    def test_rotate_y(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('y'),
            'UUUUUUUUUBBBBBBBBBRRRRRRRRRDDDDDDDDDFFFFFFFFFLLLLLLLLL',
        )

        self.assertEqual(
            cube.rotate("y'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('y2'),
            'UUUUUUUUULLLLLLLLLBBBBBBBBBDDDDDDDDDRRRRRRRRRFFFFFFFFF',
        )

    def test_rotate_z(self) -> None:
        cube = VCube()

        self.assertEqual(
            cube.rotate('z'),
            'LLLLLLLLLUUUUUUUUUFFFFFFFFFRRRRRRRRRDDDDDDDDDBBBBBBBBB',
        )

        self.assertEqual(
            cube.rotate("z'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('z2'),
            'DDDDDDDDDLLLLLLLLLFFFFFFFFFUUUUUUUUURRRRRRRRRBBBBBBBBB',
        )

    def test_rotate_invalid_modifier(self) -> None:
        cube = VCube()

        with self.assertRaises(InvalidMoveError):
            cube.rotate('z3')

    def test_rotate_invalid_move(self) -> None:
        cube = VCube()

        with self.assertRaises(InvalidMoveError):
            cube.rotate('T2')

    def test_real_case(self) -> None:
        cube = VCube()
        scramble = "U2 D2 F U2 F2 U R' L U2 R2 U' B2 D R2 L2 F2 U' L2 D F2 U'"

        self.assertEqual(
            cube.rotate(scramble),
            'FBFUUDUUDBFUFRLRRRLRLLFRRDBFBUBDBFUDRFBRLFLLULUDDBDBLD',
        )

    def test_real_case_2(self) -> None:
        cube = VCube()
        scramble = "F R' F' U' D2 B' L F U' F L' U F2 U' F2 B2 L2 D2 B2 D' L2"

        self.assertEqual(
            cube.rotate(scramble),
            'LDBRUUBBDFLUFRLBDDLURLFDFRLLFUFDRFDBFUDBLBRUURBDFBRRLU',
        )

    def test_real_case_3(self) -> None:
        cube = VCube()
        scramble = "F R F' U' D2 B' L F U' F L' U F2 U' F2 B2 L2 D2 B2 D' L2 B'"

        self.assertEqual(
            cube.rotate(scramble),
            'UFFRUUBBDFLLFRDBUFLURLFDBRLDFUBDRLLRBDDDLBFRRDURBBLUFU',
        )

    def test_real_case_with_algorithm(self) -> None:
        cube = VCube()
        scramble = parse_moves(
            "U2 D2 F U2 F2 U R' L U2 R2 U' B2 D R2 L2 F2 U' L2 D F2 U'",
        )

        self.assertEqual(
            cube.rotate(scramble),
            'FBFUUDUUDBFUFRLRRRLRLLFRRDBFBUBDBFUDRFBRLFLLULUDDBDBLD',
        )


class VCubeRotateWideSiGNTestCase(unittest.TestCase):

    def check_rotate(self, raw_move: str) -> None:
        base_move = Move(raw_move)

        for move, name in zip(
                [base_move, base_move.inverted, base_move.doubled],
                ['Base', 'Inverted', 'Doubled'],
                strict=True,
        ):
            with self.subTest(name, move=move):
                cube = VCube()
                cube_wide = VCube()

                self.assertEqual(
                    cube.rotate(str(move)),
                    cube_wide.rotate(
                        parse_moves(
                            str(move),
                        ).transform(
                            unwide_rotation_moves,
                        ),
                    ),
                )

    def test_rotate_u(self) -> None:
        self.check_rotate('u')

    def test_rotate_r(self) -> None:
        self.check_rotate('r')

    def test_rotate_f(self) -> None:
        self.check_rotate('f')

    def test_rotate_d(self) -> None:
        self.check_rotate('d')

    def test_rotate_l(self) -> None:
        self.check_rotate('l')

    def test_rotate_b(self) -> None:
        self.check_rotate('b')


class VCubeRotateWideStandardTestCase(unittest.TestCase):

    def check_rotate(self, raw_move: str) -> None:
        base_move = Move(raw_move)

        for move, name in zip(
                [base_move, base_move.inverted, base_move.doubled],
                ['Base', 'Inverted', 'Doubled'],
                strict=True,
        ):
            with self.subTest(name, move=move):
                cube = VCube()
                cube_wide = VCube()

                self.assertEqual(
                    cube.rotate(str(move)),
                    cube_wide.rotate(
                        parse_moves(
                            str(move),
                        ).transform(
                            unwide_rotation_moves,
                        ),
                    ),
                )

    def test_rotate_u(self) -> None:
        self.check_rotate('Uw')

    def test_rotate_r(self) -> None:
        self.check_rotate('Rw')

    def test_rotate_f(self) -> None:
        self.check_rotate('Fw')

    def test_rotate_d(self) -> None:
        self.check_rotate('Dw')

    def test_rotate_l(self) -> None:
        self.check_rotate('Lw')

    def test_rotate_b(self) -> None:
        self.check_rotate('Bw')


class VCubeRotateWideCancelTestCase(unittest.TestCase):

    def check_rotate(self, raw_move: str) -> None:
        base_move = Move(raw_move)

        cube = VCube()
        cube_wide = VCube()

        for move, name in zip(
                [base_move, base_move.inverted],
                ['Base', 'Inverted'],
                strict=True,
        ):
            with self.subTest(name, move=move):
                self.assertEqual(
                    cube.rotate(str(move)),
                    cube_wide.rotate(
                        parse_moves(
                            str(move),
                        ).transform(
                            unwide_rotation_moves,
                        ),
                    ),
                )

        self.assertTrue(cube_wide.is_solved)
        self.assertTrue(cube.is_solved)

    def test_rotate_u(self) -> None:
        self.check_rotate('u')

    def test_rotate_r(self) -> None:
        self.check_rotate('r')

    def test_rotate_f(self) -> None:
        self.check_rotate('f')

    def test_rotate_d(self) -> None:
        self.check_rotate('d')

    def test_rotate_l(self) -> None:
        self.check_rotate('l')

    def test_rotate_b(self) -> None:
        self.check_rotate('b')


class VCubeRotateWideDoubleCancelTestCase(unittest.TestCase):

    def check_rotate(self, raw_move: str) -> None:
        move = Move(raw_move).doubled

        cube = VCube()
        cube_wide = VCube()

        self.assertEqual(
            cube.rotate(str(move)),
            cube_wide.rotate(
                parse_moves(
                    str(move),
                ).transform(
                    unwide_rotation_moves,
                ),
            ),
        )

        self.assertEqual(
            cube.rotate(str(move)),
            cube_wide.rotate(
                parse_moves(
                    str(move),
                ).transform(
                    unwide_rotation_moves,
                ),
            ),
        )

        self.assertTrue(cube_wide.is_solved)
        self.assertTrue(cube.is_solved)

    def test_rotate_u(self) -> None:
        self.check_rotate('u')

    def test_rotate_r(self) -> None:
        self.check_rotate('r')

    def test_rotate_f(self) -> None:
        self.check_rotate('f')

    def test_rotate_d(self) -> None:
        self.check_rotate('d')

    def test_rotate_l(self) -> None:
        self.check_rotate('l')

    def test_rotate_b(self) -> None:
        self.check_rotate('b')


class VCubeRotateWideAdvancedTestCase(unittest.TestCase):

    def check_rotate(self, raw_move: str) -> None:
        base_move = Move(raw_move)

        cube = VCube()
        cube.rotate("R U R' U'")
        cube_wide = VCube()
        cube_wide.rotate("R U R' U'")

        for move, name in zip(
                [base_move, base_move.inverted],
                ['Base', 'Inverted'],
                strict=True,
        ):
            with self.subTest(name, move=move):
                self.assertEqual(
                    cube.rotate(str(move)),
                    cube_wide.rotate(
                        parse_moves(
                            str(move),
                        ).transform(
                            unwide_rotation_moves,
                        ),
                    ),
                )

    def test_rotate_u(self) -> None:
        self.check_rotate('u')

    def test_rotate_r(self) -> None:
        self.check_rotate('r')

    def test_rotate_f(self) -> None:
        self.check_rotate('f')

    def test_rotate_d(self) -> None:
        self.check_rotate('d')

    def test_rotate_l(self) -> None:
        self.check_rotate('l')

    def test_rotate_b(self) -> None:
        self.check_rotate('b')


class TestVCubeShow(unittest.TestCase):

    def setUp(self) -> None:
        self.cube = VCube()

    def test_show_default_parameters(self) -> None:
        captured_output = StringIO()
        with patch('sys.stdout', captured_output):
            self.cube.show()

        output = captured_output.getvalue()

        self.assertIsInstance(output, str)
        self.assertGreater(len(output), 0)

    def test_show_with_orientation(self) -> None:
        orientations = ['', 'DF', 'FR']

        for orientation in orientations:
            with self.subTest(orientation=orientation):
                captured_output = StringIO()
                with patch('sys.stdout', captured_output):
                    self.cube.show(orientation=orientation)

                output = captured_output.getvalue()
                self.assertIsInstance(output, str)
                self.assertGreater(len(output), 0)

    def test_show_with_mode(self) -> None:
        modes = ['f2l', 'oll', 'pll']

        for mode in modes:
            with self.subTest(mode=mode):
                captured_output = StringIO()
                with patch('sys.stdout', captured_output):
                    self.cube.show(mode=mode)

                output = captured_output.getvalue()
                self.assertIsInstance(output, str)
                self.assertGreater(len(output), 0)

    def test_show_scrambled_cube(self) -> None:
        self.cube.rotate("R U R' U'")

        captured_output = StringIO()
        with patch('sys.stdout', captured_output):
            self.cube.show()

        output = captured_output.getvalue()
        self.assertIsInstance(output, str)
        self.assertGreater(len(output), 0)

        face_letters = ['U', 'R', 'F', 'D', 'L', 'B']
        for letter in face_letters:
            self.assertEqual(output.count(letter), 9)

    def test_show_output_consistency(self) -> None:
        captured_output1 = StringIO()
        with patch('sys.stdout', captured_output1):
            self.cube.show()
        output1 = captured_output1.getvalue()

        captured_output2 = StringIO()
        with patch('sys.stdout', captured_output2):
            self.cube.show()
        output2 = captured_output2.getvalue()

        self.assertEqual(output1, output2)

    def test_show_vs_display_consistency(self) -> None:
        display_result = self.cube.display()

        captured_output = StringIO()
        with patch('sys.stdout', captured_output):
            self.cube.show()
        show_result = captured_output.getvalue()

        self.assertEqual(display_result, show_result)

    def test_show_empty_parameters(self) -> None:
        captured_output = StringIO()
        with patch('sys.stdout', captured_output):
            self.cube.show(orientation='')

        output = captured_output.getvalue()
        self.assertIsInstance(output, str)
        self.assertGreater(len(output), 0)


class TestVCubeIsEqual(unittest.TestCase):

    def setUp(self) -> None:
        self.cube1 = VCube()
        self.cube2 = VCube()

    def test_is_equal_strict_identical_cubes(self) -> None:
        self.assertTrue(self.cube1.is_equal(self.cube2, strict=True))
        self.assertTrue(self.cube1.is_equal(self.cube2))

    def test_is_equal_strict_identical_states_after_moves(self) -> None:
        self.cube1.rotate("R U R'")
        self.cube2.rotate("R U R'")
        self.assertTrue(self.cube1.is_equal(self.cube2, strict=True))

    def test_is_equal_strict_different_states(self) -> None:
        self.cube1.rotate("R U R'")
        self.cube2.rotate("L U L'")
        self.assertFalse(self.cube1.is_equal(self.cube2, strict=True))

    def test_is_equal_strict_different_orientations(self) -> None:
        self.cube1.rotate('x')  # Rotate cube
        # Both cubes are solved but have different orientations
        self.assertFalse(self.cube1.is_equal(self.cube2, strict=True))

    def test_is_equal_non_strict_identical_cubes(self) -> None:
        self.assertTrue(self.cube1.is_equal(self.cube2, strict=False))

    def test_is_equal_non_strict_same_cube_different_orientations(self) -> None:
        self.cube1.rotate('x')  # Rotate the first cube
        # Both cubes should be considered equal in non-strict mode
        self.assertTrue(self.cube1.is_equal(self.cube2, strict=False))

    def test_is_equal_non_strict_complex_orientations(self) -> None:
        # Test various rotations that should still be equal in non-strict mode
        rotations = ['x', 'y', 'z', 'x2', 'y2', 'z2', "x'", "y'", "z'"]

        for rotation in rotations:
            with self.subTest(rotation=rotation):
                cube1 = VCube()
                cube2 = VCube()
                cube1.rotate(rotation)
                self.assertTrue(cube1.is_equal(cube2, strict=False))

    def test_is_equal_non_strict_combined_rotations(self) -> None:
        self.cube1.rotate('x y z')
        self.assertTrue(self.cube1.is_equal(self.cube2, strict=False))

    def test_is_equal_non_strict_scrambled_cubes_same_pattern(self) -> None:
        scramble = "R U R' U'"
        self.cube1.rotate(scramble)
        self.cube2.rotate(scramble)

        # Both have same pattern
        self.assertTrue(self.cube1.is_equal(self.cube2, strict=True))

        # Apply different orientations
        self.cube1.rotate('x')
        self.assertTrue(self.cube1.is_equal(self.cube2, strict=False))

    def test_is_equal_non_strict_different_scrambles(self) -> None:
        self.cube1.rotate("R U R'")
        self.cube2.rotate("L U L'")
        self.assertFalse(self.cube1.is_equal(self.cube2, strict=False))

    def test_is_equal_non_strict_scramble_vs_solved(self) -> None:
        self.cube1.rotate("R U R' U'")  # Not solved
        # cube2 remains solved
        self.assertFalse(self.cube1.is_equal(self.cube2, strict=False))

    def test_is_equal_non_strict_scrambled_and_oriented(self) -> None:
        # Apply same scramble to both cubes
        scramble = "R U2 R' D' R U' R' D"
        self.cube1.rotate(scramble)
        self.cube2.rotate(scramble)

        # Orient first cube differently
        self.cube1.rotate('y x')

        # Should still be equal in non-strict mode
        self.assertTrue(self.cube1.is_equal(self.cube2, strict=False))

        # But not in strict mode
        self.assertFalse(self.cube1.is_equal(self.cube2, strict=True))

    def test_is_equal_with_invalid_states(self) -> None:
        # Test with cubes that have invalid states but same pattern
        invalid_state_list = list(INITIAL_STATE)
        invalid_state_list[4] = 'R'   # Change top center to R
        invalid_state_list[22] = 'D'  # Change front center to D
        invalid_state = ''.join(invalid_state_list)

        cube1 = VCube(invalid_state, check=False)
        cube2 = VCube(invalid_state, check=False)

        self.assertTrue(cube1.is_equal(cube2, strict=True))
        self.assertTrue(cube1.is_equal(cube2, strict=False))

    def test_is_equal_edge_case_empty_history(self) -> None:
        # Test that history doesn't affect equality
        self.cube1.rotate("R U R'", history=True)
        self.cube2.rotate("R U R'", history=False)

        self.assertTrue(self.cube1.is_equal(self.cube2, strict=True))
        self.assertTrue(self.cube1.is_equal(self.cube2, strict=False))

    def test_is_equal_reflexive_property(self) -> None:
        # A cube should always be equal to itself
        self.assertTrue(self.cube1.is_equal(self.cube1, strict=True))
        self.assertTrue(self.cube1.is_equal(self.cube1, strict=False))

        # Even after moves
        self.cube1.rotate("R U R' U'")
        self.assertTrue(self.cube1.is_equal(self.cube1, strict=True))
        self.assertTrue(self.cube1.is_equal(self.cube1, strict=False))

    def test_is_equal_symmetric_property(self) -> None:
        # If A equals B, then B equals A
        cube_oriented = VCube()
        cube_oriented.rotate('x')  # Apply orientation rotation

        test_cases = [
            (VCube(), VCube()),  # Both solved
            (VCube(), cube_oriented),  # One oriented (for non-strict)
        ]

        for cube_a, cube_b in test_cases:
            with self.subTest(cube_a=repr(cube_a), cube_b=repr(cube_b)):
                # Strict mode
                result_ab = cube_a.is_equal(cube_b, strict=True)
                result_ba = cube_b.is_equal(cube_a, strict=True)
                self.assertEqual(result_ab, result_ba)

                # Non-strict mode
                result_ab_ns = cube_a.is_equal(cube_b, strict=False)
                result_ba_ns = cube_b.is_equal(cube_a, strict=False)
                self.assertEqual(result_ab_ns, result_ba_ns)


class TestVCubeOrientation(unittest.TestCase):

    def setUp(self) -> None:
        self.cube = VCube()

    def test_orientation_solved_cube(self) -> None:
        # Solved cube should have 'UF' orientation (top=U, front=F)
        self.assertEqual(self.cube.orientation, 'UF')

    def test_orientation_after_basic_rotations(self) -> None:
        # Test each basic rotation
        expected_orientations = {
            'x': 'FD',   # x rotation: top becomes front, front becomes down
            'y': 'UR',   # y rotation: top stays, front becomes right
            'z': 'LF',   # z rotation: top becomes left, front stays
            "x'": 'BU',  # x' rotation: top becomes back, front becomes up
            "y'": 'UL',  # y' rotation: top stays, front becomes left
            "z'": 'RF',  # z' rotation: top becomes right, front stays
        }

        for move, expected in expected_orientations.items():
            with self.subTest(move=move):
                cube = VCube()
                cube.rotate(move)
                self.assertEqual(cube.orientation, expected)

    def test_orientation_after_double_rotations(self) -> None:
        expected_orientations = {
            'x2': 'DB',  # x2: top becomes down, front becomes back
            'y2': 'UB',  # y2: top stays, front becomes back
            'z2': 'DF',  # z2: top becomes down, front stays
        }

        for move, expected in expected_orientations.items():
            with self.subTest(move=move):
                cube = VCube()
                cube.rotate(move)
                self.assertEqual(cube.orientation, expected)

    def test_orientation_combined_rotations(self) -> None:
        # Test combinations of rotations
        test_cases = [
            ('x y', 'FR'),    # x then y
            ('y x', 'RD'),    # y then x
            ('z x', 'FR'),    # z then x
            ('x y z', 'DR'),  # x, y, then z
        ]

        for moves, expected in test_cases:
            with self.subTest(moves=moves):
                cube = VCube()
                cube.rotate(moves)
                self.assertEqual(cube.orientation, expected)

    def test_orientation_with_face_moves(self) -> None:
        # Basic face moves (R, U, F, D, B) shouldn't change center positions
        # Slice moves (M, E, S) are expected to change centers
        face_moves = ['R', 'U', 'F', 'D', 'B', 'L']

        for move in face_moves:
            with self.subTest(move=move):
                cube = VCube()
                original_orientation = cube.orientation
                cube.rotate(move)
                self.assertEqual(cube.orientation, original_orientation)

    def test_orientation_with_slice_moves(self) -> None:
        # Slice moves (M, E, S) are expected to change center positions
        slice_moves = {
            'M': 'BU',  # Middle slice affects centers
            'E': 'UL',  # Equatorial slice affects centers
            'S': 'LF',  # Standing slice affects centers
        }

        for move, expected in slice_moves.items():
            with self.subTest(move=move):
                cube = VCube()
                cube.rotate(move)
                self.assertEqual(cube.orientation, expected)

    def test_orientation_with_complex_sequences(self) -> None:
        # Test that face moves don't affect orientation
        # even in complex sequences
        # Using only moves that don't change centers: R, U, F, D, B L
        original_orientation = self.cube.orientation

        # Complex sequence with only face moves that preserve centers
        self.cube.rotate("R U R' U' R' F R2 U' R' U' R U R' F'")
        self.assertEqual(self.cube.orientation, original_orientation)

    def test_orientation_scrambled_cube(self) -> None:
        # Orientation should still work correctly on scrambled cubes
        self.cube.rotate("R U R' U' F R F' U2 R' U R U2")
        original_orientation = self.cube.orientation

        # Apply rotation to scrambled cube
        self.cube.rotate('x')
        self.assertNotEqual(self.cube.orientation, original_orientation)
        self.assertEqual(self.cube.orientation, 'FD')

    def test_orientation_all_24_possible_orientations(self) -> None:
        # Test all 24 possible orientations of a cube
        # Each face can be on top (6),
        # and for each top face, 4 different front faces
        orientations_found = set()

        # Generate various rotation combinations to cover all orientations
        rotation_sequences = [
            '',         # UF
            'x',        # FR
            'x2',       # DB
            "x'",       # BU
            'y',        # UL
            'y x',      # LU
            'y x2',     # BD
            "y x'",     # DL
            'y2',       # UB
            'y2 x',     # BR
            'y2 x2',    # DF
            "y2 x'",    # FD
            "y'",       # UR
            "y' x",     # RD
            "y' x2",    # BF
            "y' x'",    # FB
            'z',        # LF
            'z x',      # FU
            'z x2',     # RB
            "z x'",     # BL
            "z'",       # RF
            "z' x",     # FD
            "z' x2",    # LB
            "z' x'",    # BR
        ]

        for moves in rotation_sequences:
            cube = VCube()
            if moves:
                cube.rotate(moves)
            orientation = cube.orientation
            orientations_found.add(orientation)

        # Should find multiple unique orientations
        self.assertEqual(len(orientations_found), 24)

    def test_orientation_consistency_with_oriented_copy(self) -> None:
        # Test that orientation property is consistent with oriented_copy method
        target_orientations = ['UF', 'DF', 'FR', 'BL', 'UL', 'DR']

        for target in target_orientations:
            with self.subTest(target=target):
                cube = VCube()
                oriented_cube = cube.oriented_copy(target)
                self.assertEqual(oriented_cube.orientation, target)

    def test_orientation_with_invalid_state(self) -> None:
        # Test orientation with an unchecked/invalid state
        # Create a state with modified centers
        invalid_state_list = list(INITIAL_STATE)
        invalid_state_list[4] = 'R'   # Change top center to R
        invalid_state_list[22] = 'D'  # Change front center to D
        invalid_state = ''.join(invalid_state_list)

        cube = VCube(invalid_state, check=False)
        self.assertEqual(cube.orientation, 'RD')

    def test_orientation_property_type(self) -> None:
        # Test that orientation always returns a string
        self.assertIsInstance(self.cube.orientation, str)

        # Should always be exactly 2 characters
        self.assertEqual(len(self.cube.orientation), 2)

        # After moves, still 2 characters
        self.cube.rotate("x y z R U R'")
        self.assertIsInstance(self.cube.orientation, str)
        self.assertEqual(len(self.cube.orientation), 2)

    def test_orientation_valid_face_characters(self) -> None:
        # Orientation should only contain valid face characters
        rotations = ['', 'x', 'y', 'z', 'x2', 'y2', 'z2', 'x y', 'z x y']

        for rotation in rotations:
            with self.subTest(rotation=rotation):
                cube = VCube()
                if rotation:
                    cube.rotate(rotation)

                orientation = cube.orientation
                self.assertTrue(
                    all(char in FACES for char in orientation),
                )

    def test_orientation_specific_positions(self) -> None:
        # Test that orientation correctly reads positions 4 and 21
        cube = VCube()

        # Verify initial state
        self.assertEqual(cube.state[4], 'U')   # Top center
        self.assertEqual(cube.state[21], 'F')  # Front center
        self.assertEqual(cube.orientation, 'UF')

        # After x rotation
        cube.rotate('x')
        self.assertEqual(cube.state[4], 'F')   # Top center now F
        self.assertEqual(cube.state[21], 'D')  # Front center now D
        self.assertEqual(cube.orientation, 'FD')

    def test_orientation_edge_case_positions(self) -> None:
        # Test edge case: what if centers are swapped in an invalid way
        state = list(VCube().state)
        # Swap some centers to create an unusual but testable state
        state[4] = 'D'   # Top center = D
        state[22] = 'U'  # Front center = U
        state[31] = 'F'  # Bottom center = F (to maintain some validity)

        cube = VCube(''.join(state), check=False)
        self.assertEqual(cube.orientation, 'DU')



"""
Cube state integrity validation for cubing algorithms.

This module provides comprehensive validation of Rubik's cube states
to ensure they represent valid, solvable cube configurations.
Checks include permutation validity, orientation constraints,
color combinations, and mathematical consistency.
"""
# ruff: noqa: PLR6301
from cubing_algs.constants import CORNER_FACELET_MAP
from cubing_algs.constants import EDGE_FACELET_MAP
from cubing_algs.constants import FACE_ORDER
from cubing_algs.constants import OPPOSITE_FACES
from cubing_algs.exceptions import InvalidCubeStateError
from cubing_algs.exceptions import InvalidFaceError
from cubing_algs.facelets import facelets_to_cubies

CORNER_NUMBER = 8
CORNER_VALID_ORIENTATIONS = {0, 1, 2}

EDGE_NUMBER = 12
EDGE_VALID_ORIENTATIONS = {0, 1}


def count_inversions(permutation: list[int]) -> int:
    """
    Count the number of inversions in a permutation.

    An inversion occurs when a larger element appears before
    a smaller element in the sequence. This is used to determine
    permutation parity for cube state validation.
    """
    inversions = 0

    for i, val_i in enumerate(permutation):
        for val_j in permutation[i + 1:]:
            if val_i > val_j:
                inversions += 1

    return inversions


class VCubeIntegrityChecker:
    """
    Check integrity of VCube

    This is a mixin class that expects the following from subclasses:
    - size, face_size, face_number, _state attributes
    - get_face_center_indexes() method
    """
    size: int
    face_size: int
    face_number: int

    _state: str

    def get_face_center_indexes(self) -> list[str]:
        """
        Return the center facelet characters for each face.

        Must be implemented by subclass.
        """
        raise NotImplementedError

    def check_integrity(self) -> bool:
        """
        Performs comprehensive integrity checks on the cube state.
        """
        self.check_length()

        color_counts: dict[str, int] = {}
        for i in self._state:
            color_counts.setdefault(i, 0)
            color_counts[i] += 1

        self.check_characters(color_counts)
        self.check_colors(color_counts)
        self.check_centers()

        cp, co, ep, eo, so = facelets_to_cubies(self._state)

        self.check_corner_permutations(cp)
        self.check_corner_orientations(co)
        self.check_corner_sum(co)
        self.check_corner_colors(cp, co)

        self.check_edge_permutations(ep)
        self.check_edge_orientations(eo)
        self.check_edge_sum(eo)
        self.check_edge_colors(ep, eo)

        self.check_permutation_parity(cp, ep)

        self.check_center_orientations(so)

        return True

    def check_length(self) -> None:
        """
        Validates that the state string has the correct number of characters.
        """
        expected_length = self.face_number * self.face_size

        if len(self._state) != expected_length:
            msg = f'State string must be { expected_length } characters long'
            raise InvalidCubeStateError(msg)

    def check_characters(self, color_counts: dict[str, int]) -> None:
        """
        Validates that only valid face characters are used in the state.
        """
        if set(color_counts.keys()) - set(FACE_ORDER):
            msg = (
                'State string can only '
                f'contains { " ".join(FACE_ORDER) } characters'
            )
            raise InvalidCubeStateError(msg)

    def check_colors(self, color_counts: dict[str, int]) -> None:
        """
        Validates that each color appears exactly the expected number of times.
        """
        if not all(count == self.face_size for count in color_counts.values()):
            msg = f'State string must have { self.face_size } of each color'
            raise InvalidCubeStateError(msg)

    def check_centers(self) -> None:
        """
        Validates that all face centers are unique and properly positioned.
        """
        actual_centers = set(self.get_face_center_indexes())

        if len(actual_centers) != self.face_number:
            msg = 'Face centers must be unique'
            raise InvalidCubeStateError(msg)

    def check_corner_permutations(self, cp: list[int]) -> None:
        """
        Validates corner permutation contains exactly one of each corner piece.
        """
        if len(cp) != CORNER_NUMBER or set(cp) != set(range(CORNER_NUMBER)):
            msg = (
                'Corner permutation must contain exactly '
                'one instance of each corner (0-7)'
            )
            raise InvalidCubeStateError(msg)

    def check_corner_orientations(self, co: list[int]) -> None:
        """
        Validates corner orientations are all valid values (0, 1, or 2).
        """
        if len(co) != CORNER_NUMBER or any(
                orientation not in CORNER_VALID_ORIENTATIONS
                for orientation in co
        ):
            msg = 'Corner orientation must be 0, 1, or 2 for each corner'
            raise InvalidCubeStateError(msg)

    def check_corner_sum(self, co: list[int]) -> None:
        """
        Validates corner orientation sum is divisible by 3.
        """
        if sum(co) % 3:
            msg = 'Sum of corner orientations must be divisible by 3'
            raise InvalidCubeStateError(msg)

    def check_corner_colors(self, cp: list[int], co: list[int]) -> None:
        """
        Validates corner pieces have valid color combinations.
        """
        for i, (corner_pos, _corner_ori) in enumerate(zip(cp, co, strict=True)):
            corner_facelets = [
                self._state[facelet]
                for facelet in CORNER_FACELET_MAP[corner_pos]
            ]
            if len(set(corner_facelets)) != 3:
                msg = (
                    f'Corner { i } must have 3 different colors, '
                    f'got { corner_facelets }'
                )
                raise InvalidCubeStateError(msg)

            for j, color1 in enumerate(corner_facelets):
                for _, color2 in enumerate(corner_facelets[j + 1:], j + 1):
                    if (
                            color1 in OPPOSITE_FACES
                            and OPPOSITE_FACES[color1] == color2
                    ):
                        msg = (
                            f'Corner { i } cannot have opposite colors '
                            f'{ color1 } and { color2 }'
                        )
                        raise InvalidCubeStateError(msg)

    def check_edge_permutations(self, ep: list[int]) -> None:
        """
        Validates edge permutation contains exactly one of each edge piece.
        """
        if len(ep) != EDGE_NUMBER or set(ep) != set(range(EDGE_NUMBER)):
            msg = (
                'Edge permutation must contain exactly '
                'one instance of each edge (0-11)'
            )
            raise InvalidCubeStateError(msg)

    def check_edge_orientations(self, eo: list[int]) -> None:
        """
        Validates edge orientations are all valid values (0 or 1).
        """
        if len(eo) != EDGE_NUMBER or any(
                orientation not in EDGE_VALID_ORIENTATIONS
                for orientation in eo
        ):
            msg = 'Edge orientation must be 0 or 1 for each edge'
            raise InvalidCubeStateError(msg)

    def check_edge_sum(self, eo: list[int]) -> None:
        """
        Validates edge orientation sum is even.
        """
        if sum(eo) % 2:
            msg = 'Sum of edge orientations must be even'
            raise InvalidCubeStateError(msg)

    def check_edge_colors(self, ep: list[int], eo: list[int]) -> None:
        """
        Validates edge pieces have valid color combinations.
        """
        for i, (edge_pos, _edge_ori) in enumerate(zip(ep, eo, strict=True)):
            edge_facelets = [
                self._state[facelet]
                for facelet in EDGE_FACELET_MAP[edge_pos]
            ]
            if len(set(edge_facelets)) != 2:
                msg = (
                    f'Edge { i } must have 2 different colors, '
                    f'got { edge_facelets }'
                )
                raise InvalidCubeStateError(msg)

            color1, color2 = edge_facelets
            if color1 in OPPOSITE_FACES and OPPOSITE_FACES[color1] == color2:
                msg = (
                    f'Edge { i } cannot have opposite colors '
                    f'{ color1 } and { color2 }'
                )
                raise InvalidCubeStateError(msg)

    def check_permutation_parity(self, cp: list[int], ep: list[int]) -> None:
        """
        Validates corner and edge permutation parities match.
        """
        corner_parity = count_inversions(cp) % 2
        edge_parity = count_inversions(ep) % 2

        if corner_parity != edge_parity:
            msg = 'Corner and edge permutation parities must be equal'
            raise InvalidCubeStateError(msg)

    def check_center_orientations(self, so: list[int]) -> None:
        """
        Validates center orientations are all valid values.
        """
        valid_orientations = set(range(self.face_number))

        if len(so) != self.face_number or any(
                orientation not in valid_orientations
                for orientation in so
        ):
            msg = (
                'Center orientation must be between 0 and 5 '
                'for each center'
            )
            raise InvalidCubeStateError(msg)

    def check_face_orientations(self, faces: str) -> tuple[str, str]:
        """
        Validates and parses face orientation specification.
        """
        if not faces:
            msg = 'Specify at leat one face to orient'
            raise InvalidFaceError(msg)

        if len(faces) > 2:
            msg = f'Too much faces ({ len(faces) })'
            raise InvalidFaceError(msg)

        top_face = faces[0]
        front_face = (len(faces) > 1 and faces[1]) or ''

        if top_face not in OPPOSITE_FACES:
            msg = f'{ top_face } is an invalid face'
            raise InvalidFaceError(msg)

        if OPPOSITE_FACES[top_face] == front_face:
            msg = f'{ top_face } { front_face } are opposed faces'
            raise InvalidFaceError(msg)

        if front_face and front_face not in OPPOSITE_FACES:
            msg = f'{ front_face } is an invalid face'
            raise InvalidFaceError(msg)

        return top_face, front_face



"""
Ergonomics analysis tools for Rubik's cube algorithms.

This module provides functions to analyze the ergonomic properties
of algorithms, including hand balance, fingertrick difficulty,
regrip requirements, and overall execution comfort.
"""
from typing import TYPE_CHECKING
from typing import NamedTuple

from cubing_algs.move import Move

if TYPE_CHECKING:
    from cubing_algs.algorithm import Algorithm  # pragma: no cover


class ErgonomicsData(NamedTuple):
    """
    Container for ergonomics computation results.
    """
    total_moves: int

    # Hand balance metrics
    right_hand_moves: int
    left_hand_moves: int
    both_hand_moves: int
    hand_balance_ratio: float

    # Difficulty metrics
    regrip_count: int
    awkward_moves: int
    flow_breaks: int

    # Execution metrics
    estimated_execution_time: float
    fingertrick_difficulty: float

    # Move type distribution
    thumb_moves: int
    index_finger_moves: int
    middle_finger_moves: int
    ring_finger_moves: int

    # Comfort metrics
    comfort_score: float
    ergonomic_rating: str


# Hand assignment for different move types based on speedcubing conventions
HAND_ASSIGNMENTS = {
    # Right hand dominant moves
    'R': 'right', "R'": 'right', 'R2': 'right',
    'Rw': 'right', "Rw'": 'right', 'Rw2': 'right',
    'r': 'right', "r'": 'right', 'r2': 'right',
    'F': 'right', "F'": 'right', 'F2': 'right',
    'Fw': 'right', "Fw'": 'right', 'Fw2': 'right',
    'f': 'right', "f'": 'right', 'f2': 'right',

    # Left hand dominant moves
    'L': 'left', "L'": 'left', 'L2': 'left',
    'Lw': 'left', "Lw'": 'left', 'Lw2': 'left',
    'l': 'left', "l'": 'left', 'l2': 'left',
    'B': 'left', "B'": 'left', 'B2': 'left',
    'Bw': 'left', "Bw'": 'left', 'Bw2': 'left',
    'b': 'left', "b'": 'left', 'b2': 'left',

    # Both hands (U and D moves)
    'U': 'both', "U'": 'both', 'U2': 'both',
    'Uw': 'both', "Uw'": 'both', 'Uw2': 'both',
    'u': 'both', "u'": 'both', 'u2': 'both',
    'D': 'both', "D'": 'both', 'D2': 'both',
    'Dw': 'both', "Dw'": 'both', 'Dw2': 'both',
    'd': 'both', "d'": 'both', 'd2': 'both',

    # Slice moves (typically awkward)
    'M': 'both', "M'": 'both', 'M2': 'both',
    'E': 'both', "E'": 'both', 'E2': 'both',
    'S': 'both', "S'": 'both', 'S2': 'both',

    'x': 'both', "x'": 'both', 'x2': 'both',
    'y': 'both', "y'": 'both', 'y2': 'both',
    'z': 'both', "z'": 'both', 'z2': 'both',
}

# Finger assignments for different moves
FINGER_ASSIGNMENTS = {
    # Thumb moves (generally comfortable)
    'R': 'thumb', "R'": 'thumb', 'R2': 'thumb',
    'L': 'thumb', "L'": 'thumb', 'L2': 'thumb',

    # Index finger moves (most common U moves)
    'U': 'index', "U'": 'index', 'U2': 'index',
    'D': 'index', "D'": 'index', 'D2': 'index',

    # Middle finger moves
    'F': 'middle', "F'": 'middle', 'F2': 'middle',
    'B': 'middle', "B'": 'middle', 'B2': 'middle',

    # Ring finger moves (slice moves, generally awkward)
    'M': 'ring', "M'": 'ring', 'M2': 'ring',
    'E': 'ring', "E'": 'ring', 'E2': 'ring',
    'S': 'ring', "S'": 'ring', 'S2': 'ring',
}

# Difficulty ratings for different move types
MOVE_DIFFICULTY = {
    # Easy moves
    'R': 1.0, "R'": 1.0, 'R2': 1.2,
    'L': 1.0, "L'": 1.0, 'L2': 1.2,
    'U': 1.0, "U'": 1.0, 'U2': 1.1,
    'D': 1.1, "D'": 1.1, 'D2': 1.3,
    'F': 1.2, "F'": 1.2, 'F2': 1.4,
    'B': 1.3, "B'": 1.3, 'B2': 1.5,

    # Wide moves (slightly harder)
    'Rw': 1.1, "Rw'": 1.1, 'Rw2': 1.3,
    'Lw': 1.1, "Lw'": 1.1, 'Lw2': 1.3,
    'Uw': 1.1, "Uw'": 1.1, 'Uw2': 1.2,
    'Dw': 1.2, "Dw'": 1.2, 'Dw2': 1.4,
    'Fw': 1.3, "Fw'": 1.3, 'Fw2': 1.5,
    'Bw': 1.4, "Bw'": 1.4, 'Bw2': 1.6,

    # Slice moves (awkward)
    'M': 1.5, "M'": 1.5, 'M2': 1.8,
    'E': 1.6, "E'": 1.6, 'E2': 1.9,
    'S': 1.7, "S'": 1.7, 'S2': 2.0,

    'x': 1.0, "x'": 1.0, 'x2': 1.1,
    'y': 1.0, "y'": 1.0, 'y2': 1.1,
    'z': 1.1, "z'": 1.1, 'z2': 1.2,
}

# Threshold for considering a move awkward
AWKWARD_THRESHOLD = 1.4

# Moves that typically require regrips
REGRIP_MOVES = {
    'B', "B'", 'B2', 'Bw', "Bw'", 'Bw2', 'b', "b'", 'b2',
    'D', "D'", 'D2', 'Dw', "Dw'", 'Dw2', 'd', "d'", 'd2',
    'E', "E'", 'E2', 'S', "S'", 'S2',
}

# Move pairs that create flow breaks (awkward transitions)
AWKWARD_TRANSITIONS = {
    ('R', 'L'), ('L', 'R'), ("R'", "L'"), ("L'", "R'"),
    ('F', 'B'), ('B', 'F'), ("F'", "B'"), ("B'", "F'"),
    ('M', 'R'), ('R', 'M'), ("M'", "R'"), ("R'", "M'"),
    ('M', 'L'), ('L', 'M'), ("M'", "L'"), ("L'", "M'"),
}


def get_move_key(move: Move) -> str:
    """
    Get the standardized key for move lookup.

    Handles SiGN notation and layered moves by converting them
    to their base equivalents.
    """
    if move.is_pause or move.is_rotation_move:
        return str(move)

    # Convert SiGN notation to standard
    if move.is_sign_move:
        move = move.to_standard

    # Get unlayered version for lookup
    base_move = move.unlayered

    return str(base_move)


def compute_hand_balance(moves: 'Algorithm') -> tuple[int, int, int, float]:
    """
    Calculate hand balance metrics for the algorithm.
    """
    right_count = 0
    left_count = 0
    both_count = 0

    for move in moves:
        if move.is_pause:
            continue

        move_key = get_move_key(move)
        hand = HAND_ASSIGNMENTS.get(move_key, 'both')

        if hand == 'right':
            right_count += 1
        elif hand == 'left':
            left_count += 1
        else:
            both_count += 1

    # Calculate balance ratio
    # (0.5 is perfect balance, closer to 0 or 1 is imbalanced)
    total_handed = right_count + left_count
    if total_handed == 0:
        balance_ratio = 0.5
    else:
        balance_ratio = min(right_count, left_count) / total_handed

    return right_count, left_count, both_count, balance_ratio


def compute_finger_distribution(moves: 'Algorithm') -> tuple[
        int, int, int, int]:
    """
    Calculate finger usage distribution for the algorithm.
    """
    thumb_count = 0
    index_count = 0
    middle_count = 0
    ring_count = 0

    for move in moves:
        if move.is_pause:
            continue

        move_key = get_move_key(move)
        finger = FINGER_ASSIGNMENTS.get(move_key, 'index')  # Default to index

        if finger == 'thumb':
            thumb_count += 1
        elif finger == 'index':
            index_count += 1
        elif finger == 'middle':
            middle_count += 1
        elif finger == 'ring':
            ring_count += 1

    return thumb_count, index_count, middle_count, ring_count


def compute_regrip_count(moves: 'Algorithm') -> int:
    """
    Estimate the number of regrips required for the algorithm.

    Based on moves that typically require grip changes.
    """
    regrip_count = 0

    for move in moves:
        if move.is_pause:
            continue

        move_key = get_move_key(move)
        if any(regrip_move in move_key for regrip_move in REGRIP_MOVES):
            regrip_count += 1

    return regrip_count


def compute_flow_breaks(moves: 'Algorithm') -> int:
    """
    Count awkward transitions that break the flow of execution.

    Based on move sequences that are difficult to execute smoothly.
    """
    if len(moves) < 2:
        return 0

    flow_breaks = 0
    prev_move = None

    for move in moves:
        if move.is_pause:
            continue

        if prev_move is not None:
            # Get base move only
            prev_key = get_move_key(prev_move).split("'")[0].split('2')[0]
            curr_key = get_move_key(move).split("'")[0].split('2')[0]

            if (prev_key, curr_key) in AWKWARD_TRANSITIONS:
                flow_breaks += 1

        prev_move = move

    return flow_breaks


def compute_fingertrick_difficulty(moves: 'Algorithm') -> float:
    """
    Calculate overall fingertrick difficulty score.

    Based on individual move difficulties and sequence complexity.
    """
    if not moves:
        return 0.0

    total_difficulty = 0.0
    move_count = 0

    for move in moves:
        if move.is_pause:
            continue

        move_key = get_move_key(move)
        difficulty = MOVE_DIFFICULTY.get(move_key, 1.0)
        total_difficulty += difficulty
        move_count += 1

    return total_difficulty / move_count if move_count > 0 else 0.0


def compute_estimated_execution_time(moves: 'Algorithm',
                                     regrip_count: int) -> float:
    """
    Estimate algorithm execution time in seconds.

    Based on average move times and regrip penalties.
    """
    if not moves:
        return 0.0

    # Base execution times (in seconds)
    base_move_time = 0.15  # Average time per move for experienced speedcuber
    regrip_penalty = 0.07  # Additional time per regrip

    non_pause_moves = sum(1 for move in moves if not move.is_pause)

    return (non_pause_moves * base_move_time) + (regrip_count * regrip_penalty)


def compute_comfort_score(
    hand_balance_ratio: float,
    fingertrick_difficulty: float,
    regrip_count: int,
    flow_breaks: int,
    total_moves: int,
) -> float:
    """
    Calculate overall comfort score (0-100, higher is better).

    Combines various ergonomic factors into a single score.
    """
    if total_moves == 0:
        return 100.0

    # Hand balance component (0-25 points)
    balance_score = hand_balance_ratio * 25

    # Difficulty component (0-25 points, inverted)
    max_difficulty = 2.0  # Maximum expected difficulty
    difficulty_score = max(
        0, 25 - (fingertrick_difficulty / max_difficulty * 25),
    )

    # Regrip component (0-25 points, fewer regrips is better)
    regrip_ratio = min(1.0, regrip_count / total_moves)
    regrip_score = max(0, 25 - (regrip_ratio * 25))

    # Flow component (0-25 points, fewer breaks is better)
    flow_ratio = min(1.0, flow_breaks / max(1, total_moves - 1))
    flow_score = max(0, 25 - (flow_ratio * 25))

    return balance_score + difficulty_score + regrip_score + flow_score


def get_ergonomic_rating(comfort_score: float) -> str:
    """
    Convert comfort score to ergonomic rating.
    """
    if comfort_score >= 80:
        return 'Excellent'
    if comfort_score >= 65:
        return 'Good'
    if comfort_score >= 50:
        return 'Fair'
    if comfort_score >= 35:
        return 'Poor'
    return 'Very Poor'


def compute_ergonomics(algorithm: 'Algorithm') -> ErgonomicsData:  # noqa: PLR0914
    """
    Compute comprehensive ergonomics metrics for an algorithm.

    This function analyzes various aspects of algorithm ergonomics including:
    - Hand balance and coordination requirements
    - Fingertrick difficulty and execution comfort
    - Regrip requirements and flow interruptions
    - Estimated execution time and overall comfort rating

    Returns:
        ErgonomicsData: Namedtuple containing all calculated ergonomic metrics:
            - total_moves: Total number of moves (excluding pauses)
            - right_hand_moves: Number of right-hand dominant moves
            - left_hand_moves: Number of left-hand dominant moves
            - both_hand_moves: Number of moves requiring both hands
            - hand_balance_ratio: Hand balance (0.5 is perfect,
              closer to 0/1 is imbalanced)
            - regrip_count: Estimated number of regrips required
            - awkward_moves: Number of inherently difficult moves
            - flow_breaks: Number of awkward transitions between moves
            - estimated_execution_time: Estimated time to execute (seconds)
            - fingertrick_difficulty: Average difficulty score per move
            - thumb_moves: Moves primarily using thumb
            - index_finger_moves: Moves primarily using index finger
            - middle_finger_moves: Moves primarily using middle finger
            - ring_finger_moves: Moves primarily using ring finger
            - comfort_score: Overall comfort rating (0-100)
            - ergonomic_rating: Qualitative comfort assessment
    """
    # Filter out pauses for most calculations
    non_pause_moves = [move for move in algorithm if not move.is_pause]
    total_moves = len(non_pause_moves)

    if total_moves == 0:
        return ErgonomicsData(
            total_moves=0,
            right_hand_moves=0,
            left_hand_moves=0,
            both_hand_moves=0,
            hand_balance_ratio=0.5,
            regrip_count=0,
            awkward_moves=0,
            flow_breaks=0,
            estimated_execution_time=0.0,
            fingertrick_difficulty=0.0,
            thumb_moves=0,
            index_finger_moves=0,
            middle_finger_moves=0,
            ring_finger_moves=0,
            comfort_score=100.0,
            ergonomic_rating='Excellent',
        )

    # Calculate hand balance
    right_hand, left_hand, both_hand, balance_ratio = compute_hand_balance(
        algorithm,
    )

    # Calculate finger distribution
    thumb, index, middle, ring = compute_finger_distribution(algorithm)

    # Calculate difficulty metrics
    regrip_count = compute_regrip_count(algorithm)
    flow_breaks = compute_flow_breaks(algorithm)
    fingertrick_difficulty = compute_fingertrick_difficulty(algorithm)

    # Count awkward moves (those with high difficulty scores)
    awkward_moves = sum(
        1 for move in algorithm
        if not move.is_pause
        and MOVE_DIFFICULTY.get(get_move_key(move), 1.0) > AWKWARD_THRESHOLD
    )

    # Calculate execution time
    execution_time = compute_estimated_execution_time(algorithm, regrip_count)

    # Calculate overall comfort score
    comfort_score = compute_comfort_score(
        balance_ratio,
        fingertrick_difficulty,
        regrip_count,
        flow_breaks,
        total_moves,
    )

    # Get qualitative rating
    ergonomic_rating = get_ergonomic_rating(comfort_score)

    return ErgonomicsData(
        total_moves=total_moves,
        right_hand_moves=right_hand,
        left_hand_moves=left_hand,
        both_hand_moves=both_hand,
        hand_balance_ratio=balance_ratio,
        regrip_count=regrip_count,
        awkward_moves=awkward_moves,
        flow_breaks=flow_breaks,
        estimated_execution_time=execution_time,
        fingertrick_difficulty=fingertrick_difficulty,
        thumb_moves=thumb,
        index_finger_moves=index,
        middle_finger_moves=middle,
        ring_finger_moves=ring,
        comfort_score=comfort_score,
        ergonomic_rating=ergonomic_rating,
    )



from typing import TypedDict

from cubing_algs.constants import FACE_ORDER

LOADED_PALETTES: dict[str, dict[str, str]] = {}


class PaletteConfig(TypedDict, total=False):
    """
    Configuration structure for cube face color palettes.

    Defines the complete color scheme for a cube display, including face colors,
    font settings, and various background states used in different contexts.
    """
    faces: tuple[str | dict[str, str], ...]
    font: str
    masked_background: str
    adjacent_background: str
    hidden_ansi: str


def hex_to_rgb(hex_color: str) -> tuple[int, int, int]:
    """Convert hexadecimal color string to RGB tuple."""
    hex_color = hex_color.lstrip('#')

    if len(hex_color) == 3:
        hex_color = ''.join(c * 2 for c in hex_color)

    if len(hex_color) != 6:
        msg = f'Invalid hex color format: { hex_color }'
        raise ValueError(msg)

    try:
        r = int(hex_color[0:2], 16)
        g = int(hex_color[2:4], 16)
        b = int(hex_color[4:6], 16)
    except ValueError as e:
        msg = f'Invalid hex color format: { hex_color }'
        raise ValueError(msg) from e

    return (r, g, b)


def hex_to_ansi(domain: str, hex_color: str) -> str:
    """Convert hexadecimal color value to ANSI escape code."""
    r, g, b = hex_to_rgb(hex_color)
    return f'\x1b[{ domain };2;{ r };{ g };{ b }m'


def background_hex_to_ansi(hex_color: str) -> str:
    """Convert hexadecimal color value to ANSI background color code."""
    return hex_to_ansi('48', hex_color)


def foreground_hex_to_ansi(hex_color: str) -> str:
    """Convert hexadecimal color value to ANSI foreground color code."""
    return hex_to_ansi('38', hex_color)


def build_ansi_color(background_hex: str, foreground_hex: str) -> str:
    """Build a complete ANSI escape sequence with background and foreground."""
    return (
        background_hex_to_ansi(background_hex)
        + foreground_hex_to_ansi(foreground_hex)
    )


PALETTES: dict[str, PaletteConfig] = {
    'default': {
        'faces': (
            '#F5F5F5',
            '#FF0000',
            '#00D700',
            '#FFFF00',
            {
                'background': '#FF8700',
                'font_masked': '#FFAA00',
            },
            {
                'background': '#0000FF',
                'font': '#FFFFFF',
                'font_masked': '#007FFF',
                'font_adjacent': '#007FFF',
            },
        ),
    },
    'rgb': {
        'faces': (
            '#FFFFFF',
            '#FF0000',
            '#00FF00',
            '#FFFF00',
            '#FF7F00',
            '#0000FF',
        ),
        'font': '#000000',
        'masked_background': '#000000',
        'adjacent_background': '#7F7F7F',
        'hidden_ansi': build_ansi_color('#000', '#fff'),
    },
    'vibrant': {
        'faces': (
            '#FFFFFF',
            '#FF4136',
            '#2ED573',
            '#FFEAA7',
            '#FF9F43',
            '#74B9FF',
        ),
    },
    'neon': {
        'faces': (
            '#FFFFFF',
            '#FF1493',
            '#00FF7F',
            '#FFFF00',
            '#FF8C00',
            '#00BFFF',
        ),
    },
    'metal': {
        'faces': (
            '#DCDCDC',
            '#B4643C',
            '#788C50',
            '#C8A032',
            '#C87850',
            '#64A0C8',
        ),
        'adjacent_background': '#3F3F3F',
    },
    'pastel': {
        'faces': (
            '#FFFFFF',
            '#FFB6C1',
            '#98FB98',
            '#FFF192',
            '#FFDAB9',
            '#ADD8E6',
        ),
        'adjacent_background': '#666694',
    },
    'retro': {
        'faces': (
            '#FFF8DC',
            '#CC6666',
            '#90EE90',
            '#FFFF9A',
            '#FFA500',
            '#87CEFA',
        ),
        'adjacent_background': '#5555A4',
    },
    'minecraft': {
        'faces': (
            {
                'background': '#717171',
                'font_masked': '#C8C8C8',
            },
            '#E51A02',
            '#8FCA5C',
            {
                'background': '#FAE544',
                'font': '#3C3C3C',
            },
            {
                'background': '#854F2B',
                'font_masked': '#D59F7B',
            },
            '#81ACFF',
        ),
        'font': '#FFFFFF',
        'masked_background': '#575757',
        'adjacent_background': '#000016',
        'hidden_ansi': build_ansi_color('#363636', '#DDDDDD'),
    },
    'colorblind': {
        'faces': (
            '#FFFFFF',
            '#FF2120',
            '#21FF90',
            {
                'background': '#000',
                'font': '#FFF',
                'font_masked': '#C8C8C8',
            },
            '#FFA1FF',
            {
                'background': '#1A1BFF',
                'font': '#FFF',
            },
        ),
        'font': '#000',
        'masked_background': '#202020',
        'adjacent_background': '#4B0092',
        'hidden_ansi': build_ansi_color('#5D3A9B', '#FFFFFF'),
    },
    # Known palettes
    'dracula': {
        'faces': (
            '#F8F8F2',
            '#FF5555',
            '#50FA7B',
            '#F1FA8C',
            '#FFB86C',
            '#8BE9FD',
        ),
        'font': '#282A36',
        'masked_background': '#44475A',
        'adjacent_background': '#6272A4',
        'hidden_ansi': build_ansi_color('#282A36', '#F8F8F2'),
    },
    'alucard': {
        'faces': (
            {
                'background': '#FFFBEB',
                'font': '#1F1F1F',
            },
            '#CB3A2A',
            '#14710A',
            '#846E15',
            '#A34D14',
            '#036A96',
        ),
        'font': '#FFFBEB',
        'masked_background': '#1F1F1F',
        'adjacent_background': '#CFCFDE',
        'hidden_ansi': build_ansi_color('#6C664B', '#FFFBEB'),
    },
    'solarized-dark': {
        'faces': (
            {
                'background': '#073642',
                'font_masked': '#FDF6E3',
                'font_adjacent': '#FDF6E3',
            },
            '#DC322F',
            '#859900',
            '#B58900',
            '#CB4B16',
            '#268BD2',
        ),
        'font': '#FDF6E3',
        'masked_background': '#002B36',
        'adjacent_background': '#073642',
        'hidden_ansi': build_ansi_color('#002B36', '#839496'),
    },
    'solarized-light': {
        'faces': (
            {
                'background': '#EEE8D5',
                'font_masked': '#002B36',
                'font_adjacent': '#FDF6E3',
            },
            '#DC322F',
            '#859900',
            '#B58900',
            '#CB4B16',
            '#268BD2',
        ),
        'font': '#002B36',
        'masked_background': '#EEE8D5',
        'adjacent_background': '#FDF6E3',
        'hidden_ansi': build_ansi_color('#FDF6E3', '#657B83'),
    },
    # Env
    'halloween': {
        'faces': (
            '#F8F8FF',
            '#FF4500',
            '#32CD32',
            '#FFD700',
            '#FF8C00',
            {
                'background': '#483D8B',
                'font': '#DCDCDC',
                'font_masked': '#8479C7',
                'font_adjacent': '#8479C7',
            },
        ),
        'font': '#191919',
        'masked_background': '#282828',
        'hidden_ansi': build_ansi_color('#191919', '#FFA500'),
    },
    'galaxy': {
        'faces': (
            {
                'background': '#191970',
                'font_masked': '#4141FF',
                'font_adjacent': '#6666FF',
            },
            '#FF1493',
            '#8A2BE2',
            {
                'background': '#FFD700',
                'font': '#4B4B4B',
            },
            '#FF69B4',
            {
                'background': '#483D8B',
                'font_masked': '#706FD1',
                'font_adjacent': '#AAAAFF',
            },
        ),
        'font': '#FFFFFF',
        'masked_background': '#282846',
        'hidden_ansi': build_ansi_color('#141428', '#FFE1AA'),
    },
    # Dark / Goth
    'vampire': {
        'faces': (
            '#141414',
            '#181818',
            '#222222',
            '#262626',
            '#303030',
            '#343434',
        ),
        'font': '#DB0000',
        'masked_background': '#AE3030',
        'adjacent_background': '#AE6060',
        'hidden_ansi': build_ansi_color('#000000', '#AE0000'),
    },
    'ghoul': {
        'faces': (
            '#141414',
            '#181818',
            '#222222',
            '#262626',
            '#303030',
            '#343434',
        ),
        'font': '#26C4EC',
        'masked_background': '#297E97',
        'adjacent_background': '#8BD8F3',
        'hidden_ansi': build_ansi_color('#000000', '#26C4EC'),
    },
    'goblin': {
        'faces': (
            '#141414',
            '#181818',
            '#222222',
            '#262626',
            '#303030',
            '#343434',
        ),
        'font': '#00F253',
        'masked_background': '#5CA165',
        'adjacent_background': '#2AE464',
        'hidden_ansi': build_ansi_color('#000000', '#00F253'),
    },
    # Futuristic
    'cyberpunk': {
        'faces': (
            {
                'background': '#0F0F0F',
                'font_masked': '#00FFFF',
                'font_adjacent': '#00FFFF',
            },
            '#FF10F0',
            {
                'background': '#00FF96',
                'font': '#FF10F0',
            },
            {
                'background': '#FFEA00',
                'font': '#FF10F0',
            },
            '#FF4500',
            '#00BFFF',
        ),
        'font': '#00FFFF',
        'masked_background': '#2D2D2D',
        'hidden_ansi': build_ansi_color('#711C91', '#00FFFF'),
    },
    'synthwave': {
        'faces': (
            {
                'background': '#141428',
                'font_masked': '#FFFFFF',
            },
            '#FF1493',
            '#FF69B4',
            {
                'background': '#FFD700',
                'font': '#000',
            },
            '#FF4500',
            '#8A2BE2',
        ),
        'font': '#FFFFFF',
        'masked_background': '#3C3C64',
        'hidden_ansi': build_ansi_color('#282850', '#FF1493'),
    },
    'matrix': {
        'faces': (
            {
                'background': '#141414',
                'font': '#00FF00',
                'font_masked': '#009600',
                'font_adjacent': '#B8FFB8',
            },
            {
                'background': '#141414',
                'font': '#64FF64',
                'font_masked': '#009600',
                'font_adjacent': '#B8FFB8',
            },
            {
                'background': '#141414',
                'font': '#C8FFC8',
                'font_masked': '#009600',
                'font_adjacent': '#B8FFB8',
            },
            {
                'background': '#64FF64',
                'font_masked': '#009600',
                'font_adjacent': '#B8FFB8',
            },
            {
                'background': '#96FF96',
                'font_masked': '#009600',
                'font_adjacent': '#B8FFB8',
            },
            {
                'background': '#B8FFB8',
                'font_masked': '#009600',
                'font_adjacent': '#B8FFB8',
            },
        ),
        'font': '#000000',
        'masked_background': '#141414',
        'adjacent_background': '#003200',
        'hidden_ansi': build_ansi_color('#003400', '#00FF00'),
    },
    # Nature
    'sunset': {
        'faces': (
            '#FFFFFF',
            '#FF5E4D',
            '#BACB4D',
            '#FFCE54',
            '#FF7675',
            '#6C63FF',
        ),
    },
    'ocean': {
        'faces': (
            '#F0F8FF',
            '#FF6384',
            '#4BC0C0',
            '#FFCD56',
            '#72A2EB',
            '#1E90FF',
        ),
    },
    'forest': {
        'faces': (
            '#F5F5DC',
            '#DC143C',
            '#228B22',
            '#FFD700',
            {
                'background': '#8B4513',
                'font': '#C8C8C8',
                'font_masked': '#B36D45',
                'font_adjacent': '#B36D45',
            },
            {
                'background': '#191970',
                'font': '#C8C8C8',
                'font_masked': '#9B9BFF',
                'font_adjacent': '#6969FF',
            },
        ),
        'adjacent_background': '#003200',
    },
    'fire': {
        'faces': (
            '#FFFAF0',
            '#DC143C',
            '#FF4500',
            '#FFD700',
            '#FF8C00',
            '#8B0000',
        ),
        'adjacent_background': '#320000',
    },
    'ice': {
        'faces': (
            '#F0F8FF',
            '#B0C4DE',
            '#ADD8E6',
            '#E0FFFF',
            '#AFEEEE',
            '#5F9EA0',
        ),
        'adjacent_background': '#2D6CC8',
    },
    # Monochrome
    'white': {
        'faces': (
            '#FFFFFF',
            '#F8F8F8',
            '#F0F0F0',
            '#E8E8E8',
            '#E0E0E0',
            '#D8D8D8',
        ),
        'font': '#000000',
        'masked_background': '#A2A2A2',
        'adjacent_background': '#000000',
        'hidden_ansi': build_ansi_color('#FFFFFF', '#404040'),
    },
    'black': {
        'faces': (
            '#000000',
            '#202020',
            '#303030',
            '#404040',
            '#505050',
            '#606060',
        ),
        'font': '#FFFFFF',
        'masked_background': '#989898',
        'adjacent_background': '#C8C8C8',
        'hidden_ansi': build_ansi_color('#000000', '#C0C0C0'),
    },
    'red': {
        'faces': (
            '#200000',
            '#300000',
            '#400000',
            '#500000',
            '#600000',
            '#800000',
        ),
        'font': '#FFFFFF',
        'masked_background': '#AA7070',
        'adjacent_background': '#F8C8C8',
        'hidden_ansi': build_ansi_color('#000000', '#00BB00'),
    },
    'green': {
        'faces': (
            '#002000',
            '#003000',
            '#004000',
            '#005000',
            '#006000',
            '#008000',
        ),
        'font': '#FFFFFF',
        'masked_background': '#70AA70',
        'adjacent_background': '#C8F8C8',
        'hidden_ansi': build_ansi_color('#000000', '#00BB00'),
    },
    'blue': {
        'faces': (
            '#000020',
            '#000030',
            '#000040',
            '#000050',
            '#000060',
            '#000080',
        ),
        'font': '#FFFFFF',
        'masked_background': '#7070AA',
        'adjacent_background': '#C8C8F8',
        'hidden_ansi': build_ansi_color('#000000', '#0000BB'),
    },
}

DEFAULT_FONT = '#080808'

DEFAULT_MASKED_BACKGROUND = '#444444'

DEFAULT_ADJACENT_BACKGROUND = '#00004E'

DEFAULT_HIDDEN_ANSI = build_ansi_color('#303030', '#DADADA')


def build_ansi_palette(
        faces: tuple[str | dict[str, str], ...],
        font: str = DEFAULT_FONT,
        masked_background: str = DEFAULT_MASKED_BACKGROUND,
        adjacent_background: str = DEFAULT_ADJACENT_BACKGROUND,
        hidden_ansi: str = DEFAULT_HIDDEN_ANSI,
) -> dict[str, str]:
    """
    Build a complete ANSI color palette for cube face display.

    Creates ANSI escape sequences for each cube face in three states:
      normal, masked, and adjacent.

    Each face can be defined as a simple hex color string
    or a dictionary with custom background, font, and masked font colors.
    """
    palette = {
        'reset': '\x1b[0;0m',
        'hidden': hidden_ansi,
    }

    for face, face_config in zip(FACE_ORDER, faces, strict=True):
        if isinstance(face_config, dict):
            background = face_config['background']
            font_ansi = foreground_hex_to_ansi(
                face_config.get(
                    'font',
                    font,
                ),
            )
            font_masked_ansi = foreground_hex_to_ansi(
                face_config.get(
                    'font_masked',
                    background,
                ),
            )
            font_adjacent_ansi = foreground_hex_to_ansi(
                face_config.get(
                    'font_adjacent',
                    background,
                ),
            )
        else:
            background = face_config
            font_ansi = foreground_hex_to_ansi(font)
            font_masked_ansi = foreground_hex_to_ansi(background)
            font_adjacent_ansi = foreground_hex_to_ansi(background)

        ansi_face = (
            background_hex_to_ansi(background)
            + font_ansi
        )
        ansi_face_masked = (
            background_hex_to_ansi(masked_background)
            + font_masked_ansi
        )
        ansi_face_adjacent = (
            background_hex_to_ansi(adjacent_background)
            + font_adjacent_ansi
        )

        palette[face] = ansi_face
        palette[f'{ face }_masked'] = ansi_face_masked
        palette[f'{ face }_adjacent'] = ansi_face_adjacent

    return palette


def load_palette(palette_name: str) -> dict[str, str]:
    """
    Load and cache an ANSI color palette by name.

    Returns a cached palette if already loaded, otherwise builds
    and caches the palette from the PALETTES dictionary.

    Falls back to 'default' palette if the requested palette name is not found.
    """
    if palette_name not in PALETTES:
        palette_name = 'default'

    if palette_name in LOADED_PALETTES:
        return LOADED_PALETTES[palette_name]

    palette = build_ansi_palette(
        **PALETTES[palette_name],
    )

    LOADED_PALETTES[palette_name] = palette

    return palette



from cubing_algs.algorithm import Algorithm
from cubing_algs.constants import MAX_ITERATIONS
from cubing_algs.transform.optimize import optimize_do_undo_moves
from cubing_algs.transform.optimize import optimize_double_moves
from cubing_algs.transform.optimize import optimize_repeat_three_moves
from cubing_algs.transform.optimize import optimize_triple_moves


def compress_moves(
        old_moves: Algorithm,
        max_iterations: int = MAX_ITERATIONS,
) -> Algorithm:
    """
    Optimize an algorithm by applying move compression techniques.

    Repeatedly applies optimization functions to reduce the number of moves
    by eliminating redundancies and combining moves.
    """
    moves = old_moves.copy()

    for _ in range(max_iterations):
        start_length = len(moves)

        for optimizer in (
            optimize_do_undo_moves,
            optimize_repeat_three_moves,
            optimize_double_moves,
            optimize_triple_moves,
        ):
            moves = optimizer(moves)

        if len(moves) == start_length:
            break

    return moves


def expand_moves(old_moves: Algorithm) -> Algorithm:
    """
    Expand an algorithm by converting double moves to two single moves.

    Replaces each double move (like R2) with two identical single moves (R R).
    """
    moves = [
        expanded_move
        for move in old_moves
        for expanded_move in (
                (move.doubled, move.doubled)
                if move.is_double
                else (move,)
        )
    ]

    return Algorithm(moves)



import unittest

from cubing_algs.constants import INITIAL_STATE
from cubing_algs.masks import _CACHE_SIZE_LIMIT
from cubing_algs.masks import _MASK_CACHE
from cubing_algs.masks import FULL_MASK
from cubing_algs.masks import facelets_masked
from cubing_algs.masks import intersection_masks
from cubing_algs.masks import negate_mask
from cubing_algs.masks import state_masked
from cubing_algs.masks import union_masks


class TestBinaryMasks(unittest.TestCase):

    def test_union(self) -> None:
        self.assertEqual(union_masks('1010', '0110'), '1110')
        self.assertEqual(union_masks('1111', '0000'), '1111')
        self.assertEqual(union_masks('0000', '0000'), '0000')

    def test_union_multiple(self) -> None:
        self.assertEqual(union_masks('1000', '0100', '0010', '0001'), '1111')
        self.assertEqual(union_masks('1010', '0101', '1100'), '1111')

    def test_union_single(self) -> None:
        self.assertEqual(union_masks('1010'), '1010')

    def test_union_empty(self) -> None:
        self.assertEqual(union_masks(), '')

    def test_intersection(self) -> None:
        self.assertEqual(intersection_masks('1010', '1100'), '1000')
        self.assertEqual(intersection_masks('1111', '0000'), '0000')
        self.assertEqual(intersection_masks('1111', '1111'), '1111')

    def test_intersection_multiple(self) -> None:
        self.assertEqual(intersection_masks('1111', '1110', '1100'), '1100')
        self.assertEqual(intersection_masks('1010', '0110', '1100'), '0000')

    def test_intersection_single(self) -> None:
        self.assertEqual(intersection_masks('1010'), '1010')

    def test_intersection_empty(self) -> None:
        self.assertEqual(intersection_masks(), '')

    def test_negate(self) -> None:
        self.assertEqual(negate_mask('1010'), '0101')
        self.assertEqual(negate_mask('0000'), '1111')
        self.assertEqual(negate_mask('1111'), '0000')

    def test_negate_single_bit(self) -> None:
        self.assertEqual(negate_mask('1'), '0')
        self.assertEqual(negate_mask('0'), '1')

    def test_negate_empty(self) -> None:
        self.assertEqual(negate_mask(''), '')

    def test_facelets_masked_basic(self) -> None:
        facelets = 'ABCD'
        mask = '1010'
        expected = 'A-C-'
        self.assertEqual(facelets_masked(facelets, mask), expected)

    def test_facelets_masked_all_ones(self) -> None:
        facelets = 'ABCD'
        mask = '1111'
        expected = 'ABCD'
        self.assertEqual(facelets_masked(facelets, mask), expected)

    def test_facelets_masked_all_zeros(self) -> None:
        facelets = 'ABCD'
        mask = '0000'
        expected = '----'
        self.assertEqual(facelets_masked(facelets, mask), expected)

    def test_facelets_masked_empty(self) -> None:
        facelets = ''
        mask = ''
        expected = ''
        self.assertEqual(facelets_masked(facelets, mask), expected)

    def test_facelets_masked_single_char(self) -> None:
        facelets = 'X'
        mask = '1'
        expected = 'X'
        self.assertEqual(facelets_masked(facelets, mask), expected)

        facelets = 'X'
        mask = '0'
        expected = '-'
        self.assertEqual(facelets_masked(facelets, mask), expected)

    def test_facelets_masked_real_cube_pattern(self) -> None:
        facelets = INITIAL_STATE[:9]
        mask = '101010101'
        expected = 'U-U-U-U-U'
        self.assertEqual(facelets_masked(facelets, mask), expected)

    def test_state_masked_basic(self) -> None:
        mask = FULL_MASK
        result = state_masked(INITIAL_STATE, mask)

        self.assertEqual(result, INITIAL_STATE)

    def test_state_masked_all_zeros(self) -> None:
        mask = '0' * 54
        result = state_masked(INITIAL_STATE, mask)

        self.assertTrue(result.replace('0', '-'), mask)

    def test_state_masked_partial(self) -> None:
        mask = '1' * 9 + '0' * 45
        result = state_masked(INITIAL_STATE, mask)

        self.assertEqual(
            result,
            'UUUUUUUUU---------------------------------------------',
        )

    def test_state_masked_different_state(self) -> None:
        scrambled_state = (
            'LUULUUFFFLBBRRRRRRUUUFFDFFDRRBDDBDDBFFRLLDLLDLLDUBBUBB'
        )
        mask = '1' * 27 + '0' * 27
        result = state_masked(scrambled_state, mask)

        self.assertEqual(
            result,
            '-UU-UUFFF---RRRRRRUUUFF-FF-RR-------FFR---------U--U--',
        )

    def test_facelets_masked_cache_hit(self) -> None:
        """Test cache hit path in optimized facelets_masked."""
        _MASK_CACHE.clear()

        facelets = 'ABCD'
        mask = '1010'

        # First call - cache miss
        result1 = facelets_masked(facelets, mask)
        self.assertEqual(result1, 'A-C-')
        self.assertIn(mask, _MASK_CACHE)

        # Second call - cache hit (covers lines 74-75)
        result2 = facelets_masked(facelets, mask)
        self.assertEqual(result2, 'A-C-')
        self.assertEqual(result1, result2)

    def test_facelets_masked_cache_eviction(self) -> None:
        """Test cache eviction when size limit is reached."""
        _MASK_CACHE.clear()

        # Directly manipulate cache to test eviction by filling it beyond limit
        # This allows us to test the eviction logic paths (lines 86-88)

        # Fill cache manually to exactly the limit
        for i in range(_CACHE_SIZE_LIMIT):
            # Create unique keys and dummy values
            key_prefix = f'test_mask_{i:04d}'
            mask_key = key_prefix + '0' * (54 - len(key_prefix))
            _MASK_CACHE[mask_key] = tuple(bool(j % 2) for j in range(54))

        # Verify cache is at limit
        self.assertEqual(len(_MASK_CACHE), _CACHE_SIZE_LIMIT)

        # Now call facelets_masked with a new unique mask to trigger eviction
        test_facelets = INITIAL_STATE
        trigger_mask = '1' + '0' * 53  # Unique mask not in cache

        result = facelets_masked(test_facelets, trigger_mask)

        # Verify eviction occurred - cache should be reduced and new mask added
        expected_size = _CACHE_SIZE_LIMIT // 2 + 1
        self.assertEqual(len(_MASK_CACHE), expected_size)

        # The triggering mask should be in the cache
        self.assertIn(trigger_mask, _MASK_CACHE)

        # Verify the result is correct
        expected_result = 'U' + '-' * 53
        self.assertEqual(result, expected_result)

    def test_facelets_masked_cache_behavior_with_repeated_patterns(
        self,
    ) -> None:
        """Test cache behavior with realistic repeated mask usage."""
        _MASK_CACHE.clear()

        # Test with cube-sized strings and common patterns
        facelets = INITIAL_STATE
        common_masks = [FULL_MASK, '0' * 54, '1' * 27 + '0' * 27]

        # First round - populate cache
        results1 = []
        for mask in common_masks:
            results1.append(facelets_masked(facelets, mask))

        # Verify all masks are cached
        for mask in common_masks:
            self.assertIn(mask, _MASK_CACHE)

        # Second round - should hit cache
        results2 = []
        for mask in common_masks:
            results2.append(facelets_masked(facelets, mask))

        # Results should be identical
        self.assertEqual(results1, results2)



from cubing_algs.algorithm import Algorithm
from cubing_algs.constants import FACE_INDEXES
from cubing_algs.constants import FACE_ORDER
from cubing_algs.constants import INITIAL_STATE
from cubing_algs.constants import OFFSET_ORIENTATION_MAP
from cubing_algs.display import VCubeDisplay
from cubing_algs.exceptions import InvalidMoveError
from cubing_algs.extensions import rotate
from cubing_algs.facelets import cubies_to_facelets
from cubing_algs.facelets import facelets_to_cubies
from cubing_algs.integrity import VCubeIntegrityChecker
from cubing_algs.move import Move


class VCube(VCubeIntegrityChecker):
    """
    Virtual 3x3 cube for tracking moves on facelets.

    Represents a Rubik's cube state using a 54-character string
    where each character represents a facelet color.
    """
    size = 3
    face_number = 6
    face_size = size * size

    def __init__(self, initial: str | None = None, *,
                 check: bool = True,
                 history: list[str] | None = None):
        """
        Initialize a virtual cube with optional initial state and history.
        """
        if initial:
            self._state = initial
            if check:
                self.check_integrity()
        else:
            self._state = INITIAL_STATE

        self.history: list[str] = history or []

    @property
    def state(self) -> str:
        """
        Get the current state of the cube as a facelet string.
        """
        return self._state

    @staticmethod
    def from_cubies(cp: list[int], co: list[int],  # noqa: PLR0913 PLR0917
                    ep: list[int], eo: list[int],
                    so: list[int],
                    scheme: str | None = None) -> 'VCube':
        """
        Create a VCube from cubie representation.
        """
        return VCube(
            cubies_to_facelets(cp, co, ep, eo, so, scheme),
            check=not bool(scheme),
        )

    @property
    def to_cubies(self) -> tuple[
            list[int], list[int], list[int], list[int], list[int],
    ]:
        """
        Convert the cube state to cubie representation.
        """
        return facelets_to_cubies(self._state)

    @property
    def is_solved(self) -> bool:
        """
        Check if the cube is in a solved state.
        """
        return all(face * self.face_size in self.state for face in FACE_ORDER)

    def is_equal(self, other_cube: 'VCube', *, strict: bool = True) -> bool:
        """
        Compare two cubes for equality with optional orientation flexibility.

        In strict mode, compares exact facelet states.
        In non-strict mode, reorients the other cube to match
        this cube's orientation before comparing.
        """
        if strict:
            return self.state == other_cube.state

        oriented_copy = other_cube.oriented_copy(self.orientation)

        return self.state == oriented_copy.state

    @property
    def orientation(self) -> str:
        """
        Get the cube's orientation as a two-character string.

        Uses the top face center and front face center
        to determine the current orientation of the cube in space.

        It might not work well with an unchecked state.
        """
        return self._state[4] + self._state[22]

    def rotate(self, moves: Algorithm | Move | str, *,
               history: bool = True) -> str:
        """
        Apply a sequence of moves to the cube.
        """
        moves_str = str(moves)

        if not moves_str:
            return self._state

        for move in moves_str.split(' '):
            self.rotate_move(move, history=history)

        return self._state

    def rotate_move(self, move: str, *, history: bool = True) -> str:
        """
        Apply a single move to the cube.
        """
        try:
            self._state = rotate.rotate_move(self._state, move)
        except ValueError as e:
            raise InvalidMoveError(str(e)) from e
        else:
            if history:
                self.history.append(move)
            return self._state

    def copy(self, *, full: bool = False) -> 'VCube':
        """
        Create a copy of the cube with optional history preservation.
        """
        history = None
        if full:
            history = list(self.history)

        return VCube(
            self.state,
            check=False,
            history=history,
        )

    def compute_orientation_moves(self, faces: str) -> str:
        """
        Calculate the moves needed to orient the cube to specific faces.
        """
        top_face, front_face = self.check_face_orientations(faces)

        orientation_key = str(self.get_face_index(top_face))

        if front_face:
            orientation_key += str(self.get_face_index(front_face))

        return OFFSET_ORIENTATION_MAP[orientation_key]

    def oriented_copy(self, faces: str, *, full: bool = False) -> 'VCube':
        """
        Create a copy of the cube oriented to specific faces.
        """
        cube = self.copy(full=full)

        moves = self.compute_orientation_moves(faces)

        if moves:
            cube.rotate(moves, history=full)

        return cube

    def display(self, mode: str = '', orientation: str = '',  # noqa: PLR0913 PLR0917
                mask: str = '', palette: str = '',
                effect: str = '', facelet: str = '') -> str:
        """
        Generate a visual representation of the cube.
        """
        return VCubeDisplay(self, palette, effect, facelet).display(
            mode, orientation, mask,
        )

    def show(self, mode: str = '', orientation: str = '',  # noqa: PLR0913 PLR0917
             mask: str = '', palette: str = '',
             effect: str = '', facelet: str = '') -> None:
        """
        Print a visual representation of the cube.
        """
        print(
            self.display(
                mode, orientation, mask,
                palette, effect, facelet,
            ),
            end='',
        )

    def get_face(self, face: str) -> str:
        """
        Get the facelets of a specific face by face letter.
        """
        index = FACE_INDEXES[face]
        return self._state[index * self.face_size: (index + 1) * self.face_size]

    def get_face_center_indexes(self) -> list[str]:
        """
        Get the center facelet colors for all faces.
        """
        face_centers = []

        for i in range(6):
            face_centers.append(self.state[(i * self.face_size) + 4])

        return face_centers

    def get_face_index(self, face: str) -> int:
        """
        Get the index of a face by its center color.
        """
        return self.get_face_center_indexes().index(face)

    def get_face_by_center(self, face: str) -> str:
        """
        Get the facelets of a face by its center color.
        """
        index = self.get_face_index(face)

        return self._state[index * self.face_size: (index + 1) * self.face_size]

    def __str__(self) -> str:
        """
        Return the facelets of the cube
        """
        faces = []
        for face in FACE_ORDER:
            faces.append(f'{ face }: { self.get_face(face)}')

        return '\n'.join(faces)

    def __repr__(self) -> str:
        """
        Return a string representation that can be used
        to recreate the VCube.
        """
        return f"VCube('{ self._state }')"



from cubing_algs.algorithm import Algorithm
from cubing_algs.constants import MAX_ITERATIONS
from cubing_algs.constants import REWIDE_MOVES
from cubing_algs.constants import UNWIDE_ROTATION_MOVES
from cubing_algs.constants import UNWIDE_SLICE_MOVES
from cubing_algs.move import Move


def unwide(
        old_moves: Algorithm,
        config: dict[str, list[str]],
) -> Algorithm:
    """
    Expand wide moves using the provided configuration mapping.
    """
    moves: list[Move] = []

    move_cache: dict[Move, list[Move]] = {}
    for move_str, replacements in config.items():
        move_cache[Move(move_str)] = [Move(m) for m in replacements]

    for move in old_moves:
        move_untimed = move.untimed

        if move_untimed in config:
            if move.is_timed:
                moves.extend(
                    [
                        Move(x + move.time)
                        for x in move_cache[move_untimed]
                    ],
                )
            else:
                moves.extend(move_cache[move_untimed])
        else:
            moves.append(move)

    return Algorithm(moves)


def unwide_slice_moves(old_moves: Algorithm) -> Algorithm:
    """
    Expand wide moves into outer face and slice moves.
    """
    return unwide(old_moves, UNWIDE_SLICE_MOVES)


def unwide_rotation_moves(old_moves: Algorithm) -> Algorithm:
    """
    Expand wide moves into outer face and rotation moves.
    """
    return unwide(old_moves, UNWIDE_ROTATION_MOVES)


def rewide(
        old_moves: Algorithm,
        config: dict[str, str],
        max_depth: int = MAX_ITERATIONS,
) -> Algorithm:
    """
    Convert sequences of moves back into wide moves using configuration.
    """
    if max_depth <= 0:
        return old_moves

    i = 0
    moves: list[Move] = []
    changed = False

    while i < len(old_moves) - 1:
        wided = f'{ old_moves[i].untimed } { old_moves[i + 1].untimed }'
        if wided in config:
            moves.append(Move(f'{ config[wided] }{ old_moves[i].time }'))
            changed = True
            i += 2
        else:
            moves.append(old_moves[i])
            i += 1

    if i < len(old_moves):
        moves.append(old_moves[i])

    if changed:
        return rewide(
            Algorithm(moves), config,
            max_depth - 1,
        )

    return Algorithm(moves)


def rewide_moves(old_moves: Algorithm) -> Algorithm:
    """
    Convert move sequences back into wide moves where possible.
    """
    return rewide(old_moves, REWIDE_MOVES)



import unittest

from cubing_algs.metrics import MetricsData
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.optimize import optimize_double_moves


class MetricsTestCase(unittest.TestCase):
    maxDiff = None

    def test_metrics(self) -> None:
        algo = parse_moves("yM2UMU2M'UM2")
        self.assertEqual(
            algo.metrics,
            MetricsData(
                generators=['M', 'U'],
                inner_moves=4,
                outer_moves=3,
                pauses=0,
                rotations=1,
                htm=11,
                qtm=16,
                stm=7,
                etm=8,
                rtm=1,
                qstm=10,
            ),
        )

    def test_htm(self) -> None:
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 1, 2, 2, 0, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics.htm, score)

    def test_qtm(self) -> None:
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 2, 2, 4, 0, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics.qtm, score)

    def test_stm(self) -> None:
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 1, 1, 1, 0, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics.stm, score)

    def test_etm(self) -> None:
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 1, 1, 1, 1, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics.etm, score)

    def test_qstm(self) -> None:
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 2, 1, 2, 0, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics.qstm, score)

    def test_rtm(self) -> None:
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'", 'x y2']
        scores = [0, 0, 0, 0, 2, 0, 3]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics.rtm, score)

    def test_issue_11(self) -> None:
        moves = "R U F' B R' U F' U' F D F' D' F' D' F D' L D L' R D' R' D' B D' B' D' D' R D' D' R' D B' D' B D' D' F D' F' D F D F' D' D D' D' L D B D' B' L' D R F D F' D' R' R F D' F' D' F D F' R' F D F' D' F' R F R' D"  # noqa: E501

        algo = parse_moves(moves)
        self.assertEqual(
            algo.metrics,
            MetricsData(
                generators=['D', 'F', 'R', 'B', 'L', 'U'],
                inner_moves=0,
                outer_moves=80,
                pauses=0,
                rotations=0,
                htm=80,
                qtm=80,
                stm=80,
                etm=80,
                rtm=0,
                qstm=80,
            ),
        )

        compress = algo.transform(optimize_double_moves)

        self.assertEqual(
            compress.metrics,
            MetricsData(
                generators=['D', 'F', 'R', 'B', 'L', 'U'],
                inner_moves=0,
                outer_moves=76,
                pauses=0,
                rotations=0,
                htm=76,
                qtm=80,
                stm=76,
                etm=76,
                rtm=0,
                qstm=80,
            ),
        )

    def test_metrics_wide_sign(self) -> None:
        algo = parse_moves('RFu')
        self.assertEqual(
            algo.metrics,
            MetricsData(
                generators=['R', 'F', 'u'],
                inner_moves=0,
                outer_moves=3,
                pauses=0,
                rotations=0,
                htm=3,
                qtm=3,
                stm=3,
                etm=3,
                rtm=0,
                qstm=3,
            ),
        )

    def test_metrics_wide_standard(self) -> None:
        algo = parse_moves('RFUw')
        self.assertEqual(
            algo.metrics,
            MetricsData(
                generators=['R', 'F', 'Uw'],
                inner_moves=0,
                outer_moves=3,
                pauses=0,
                rotations=0,
                htm=3,
                qtm=3,
                stm=3,
                etm=3,
                rtm=0,
                qstm=3,
            ),
        )

    def test_metrics_pauses(self) -> None:
        algo = parse_moves('R..Fu.')
        self.assertEqual(
            algo.metrics,
            MetricsData(
                generators=['R', 'F', 'u'],
                inner_moves=0,
                outer_moves=3,
                pauses=3,
                rotations=0,
                htm=3,
                qtm=3,
                stm=3,
                etm=3,
                rtm=0,
                qstm=3,
            ),
        )
